#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import collections
import typing as t
from numbers import Number

from typing_extensions import Literal

from pyxel.evaluator import eval_range


class ParameterValues:
    """TBW."""

    def __init__(
        self,
        key: str,
        values: t.Union[
            Literal["_"], t.Sequence[Literal["_"]], t.Sequence[Number], t.Sequence[str]
        ],
        boundaries: t.Optional[t.Tuple[float, float]] = None,
        enabled: bool = True,
        logarithmic: bool = False,
    ):
        # TODO: should these values be checked ?
        assert values == "_" or (
            isinstance(values, collections.Sequence)
            and (
                all([el == "_" for el in values])
                or all([isinstance(el, str) for el in values])
                or all([isinstance(el, Number) for el in values])
            )
        )

        # unique identifier to the step. example: 'detector.geometry.row'
        self._key = key  # type: str
        self._values = (
            values
        )  # type: t.Union[Literal["_"], t.Sequence[Literal["_"]], t.Sequence[Number], t.Sequence[str]]

        self._enabled = enabled  # type: bool
        self._logarithmic = logarithmic  # type: bool
        self._boundaries = boundaries  # type: t.Optional[t.Tuple[float, float]]

        self._current = None  # type: t.Optional[t.Union[Literal['_'], Number, str]]

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__  # type: str

        return f"{cls_name}<key={self.key!r}, values={self.values!r}, enabled={self.enabled!r}>"

    def __len__(self) -> int:
        values = eval_range(self.values)  # type: t.Sequence
        return len(values)

    # TODO: Is method '__contains__' needed ? If yes then this class will act as a `Collections.abc.Sequence`
    def __iter__(self) -> t.Iterator[Number]:
        values = eval_range(self.values)
        for value in values:
            yield value

    @property
    def key(self) -> str:
        """TBW."""
        return self._key

    @property
    def values(
        self,
    ) -> t.Union[
        Literal["_"], t.Sequence[Literal["_"]], t.Sequence[Number], t.Sequence[str]
    ]:
        """TBW."""
        return self._values

    @property
    def enabled(self) -> bool:
        """TBW."""
        return self._enabled

    @property
    def current(self) -> t.Optional[t.Union[Literal["_"], Number, str]]:
        """TBW."""
        return self._current

    @current.setter
    def current(self, value: t.Union[Literal["_"], str, Number]) -> None:
        """TBW."""
        self._current = value

    @property
    def logarithmic(self) -> bool:
        """TBW."""
        return self._logarithmic

    @property
    def boundaries(self) -> t.Optional[t.Tuple[float, float]]:
        """TBW."""
        return self._boundaries



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel TARS model to generate charge by ionization."""

import typing as t

import numpy as np
from typing_extensions import Literal

from pyxel.detectors import Detector
from pyxel.models.charge_generation.tars.util import sampling_distribution


class Particle:
    """Particle class define a particle together with its characteristics."""

    def __init__(
        self,
        detector: Detector,
        simulation_mode: Literal[
            "cosmic_ray", "cosmics", "radioactive_decay", "snowflakes"
        ],
        particle_type: Literal[
            "proton", "ion", "alpha", "beta", "electron", "gamma", "x-ray"
        ],
        input_energy: t.Union[int, float, Literal["random"]],
        spectrum_cdf: np.ndarray,
        starting_pos_ver: t.Union[str, np.ndarray],
        starting_pos_hor: t.Union[str, np.ndarray],
        starting_pos_z: t.Union[str, np.ndarray],
        # input_alpha='random', input_beta='random'
    ):
        """Creation of a particle according to some parameters.

        :param detector:
        :param simulation_mode:
        :param particle_type:
        :param input_energy:
        :param spectrum_cdf:
        :param starting_pos_ver:
        :param starting_pos_hor:
        :param starting_pos_z:
        """
        self.detector = detector
        geo = self.detector.geometry

        # starting_position_vertical = None
        # starting_position_horizontal = None
        # starting_position_z = None

        # if starting_pos_ver == 'random':
        #     starting_position_vertical = geo.vert_dimension * np.random.random()
        # elif isinstance(starting_pos_ver, int) or isinstance(starting_pos_ver, float):
        #     starting_position_vertical = starting_pos_ver
        # if starting_pos_hor == 'random':
        #     starting_position_horizontal = geo.horz_dimension * np.random.random()
        # elif isinstance(starting_pos_hor, int) or isinstance(starting_pos_hor, float):
        #     starting_position_horizontal = starting_pos_hor
        #
        # if starting_pos_z == 'random':
        #     starting_position_z = geo.total_thickness * np.random.random()
        # elif isinstance(starting_pos_z, int) or isinstance(starting_pos_z, float):
        #     starting_position_z = starting_pos_z
        #
        # self.starting_position = np.array([starting_position_vertical,
        #                                    starting_position_horizontal,
        #                                    starting_position_z])
        # self.position = np.copy(self.starting_position)
        # self.trajectory = np.copy(self.starting_position)

        # if input_alpha == 'random' and starting_pos_z == 0.:
        #     alpha = 2 * math.pi * np.random.random()
        # elif input_alpha == 'random' and starting_pos_z != 0.:
        #     alpha = 2 * math.pi * np.random.random()
        # else:
        #     alpha = input_alpha  # between 0 and 2*pi
        #
        # if input_beta == 'random':
        #     beta = 2. * math.pi * np.random.random()
        # else:
        #     beta = input_beta
        # self.angle = np.array([alpha, beta])
        #
        # self.dir_z = -1 * math.sin(alpha)
        # self.dir_ver = math.cos(alpha) * math.cos(beta)
        # self.dir_hor = math.cos(alpha) * math.sin(beta)

        # if input_alpha != 'random':
        #     self.alpha = input_alpha
        # if input_beta != 'random':
        #     self.beta = input_beta
        # # update direction:
        # self.dir_ver, self.dir_hor, self.dir_z = get_direction_from_angles()

        self.track_length = None  # type: t.Optional[float]

        self.dir_ver, self.dir_hor, self.dir_z = isotropic_direction()

        self.random_det_pt_vert = geo.vert_dimension * np.random.random()
        self.random_det_pt_horz = geo.horz_dimension * np.random.random()
        self.random_det_pt_z = -1 * geo.total_thickness * np.random.random()

        self.alpha, self.beta = self.get_angles()  # rad

        mode_1 = ["cosmic_ray", "cosmics"]
        mode_2 = ["radioactive_decay", "snowflakes"]
        if (
            simulation_mode in mode_1
        ):  # cosmic rays coming from OUTSIDE the detector volume
            self.starting_position = self.get_surface_point()
        elif simulation_mode in mode_2:  # radioactive decay INSIDE the detector volume
            self.starting_position = np.array(
                [self.random_det_pt_vert, self.random_det_pt_horz, self.random_det_pt_z]
            )

        if starting_pos_ver != "random":
            self.starting_position[0] = starting_pos_ver
        if starting_pos_hor != "random":
            self.starting_position[1] = starting_pos_hor
        if starting_pos_z != "random":
            self.starting_position[2] = starting_pos_z

        self.position = np.copy(self.starting_position)
        self.trajectory = np.copy(self.starting_position)

        if input_energy == "random":
            self.energy = sampling_distribution(spectrum_cdf)  # type: float
        elif isinstance(input_energy, int) or isinstance(input_energy, float):
            self.energy = input_energy
        else:
            raise ValueError("Given particle energy could not be read")

        self.deposited_energy = 0.0  # type: float
        self.total_edep = 0.0  # type: float

        self.type = particle_type
        ionizing_particles = ("proton", "ion", "alpha", "beta", "electron")
        non_ionizing_particles = ("gamma", "x-ray")  # 'photon'

        if self.type in ionizing_particles:
            # call direct ionization func when needed - already implemented in simulation
            pass

        elif self.type in non_ionizing_particles:
            # call NON-direct ionization func when needed - need to be implemented
            raise NotImplementedError(
                "Given particle type simulation is not yet implemented"
            )

        else:
            raise ValueError("Given particle type can not be simulated")

    def get_surface_point(self) -> np.ndarray:
        """TBW.

        :return:
        """
        geo = self.detector.geometry

        norm_vectors = [
            np.array(
                [0.0, 0.0, -1.0]
            ),  # top plane (usually particle enters vol. via this)
            np.array(
                [0.0, 0.0, 1.0]
            ),  # bottom plane (usually particle leaves vol. via this)
            np.array([0.0, 1.0, 0.0]),
            np.array([-1.0, 0.0, 0.0]),
            np.array([0.0, -1.0, 0.0]),
            np.array([1.0, 0.0, 0.0]),
        ]  # type: t.List[np.ndarray]

        points = [
            np.array(
                [0.0, 0.0, 0.0]
            ),  # top plane (usually particle enters vol. via this)
            np.array(
                [0.0, 0.0, -1 * geo.total_thickness]
            ),  # bottom plane (usually particle leaves vol. via this)
            np.array([0.0, 0.0, 0.0]),
            np.array([geo.vert_dimension, 0.0, 0.0]),
            np.array([geo.vert_dimension, geo.horz_dimension, 0.0]),
            np.array([0.0, geo.horz_dimension, 0.0]),
        ]  # type: t.List[np.ndarray]

        intersect_points = np.zeros((6, 3))
        track_direction = np.array([self.dir_ver, self.dir_hor, self.dir_z])
        random_det_point = np.array(
            [self.random_det_pt_vert, self.random_det_pt_horz, self.random_det_pt_z]
        )

        surface_start_point = None  # type: t.Optional[np.ndarray]
        surface_end_point = None  # type: t.Optional[np.ndarray]
        for i in range(6):
            intersect_points[i, :] = find_intersection(
                n=norm_vectors[i], p0=points[i], ls=random_det_point, lv=track_direction
            )

            eps = 1e-8
            if 0.0 - eps <= intersect_points[i, 0] <= geo.vert_dimension + eps:
                if 0.0 - eps <= intersect_points[i, 1] <= geo.horz_dimension + eps:
                    if (
                        -1 * geo.total_thickness - eps
                        <= intersect_points[i, 2]
                        <= 0.0 + eps
                    ):
                        if (
                            np.dot(
                                track_direction,
                                intersect_points[i, :] - random_det_point,
                            )
                            < 0
                        ):
                            surface_start_point = self.intersection_correction(
                                intersect_points[i, :]
                            )
                        else:
                            surface_end_point = self.intersection_correction(
                                intersect_points[i, :]
                            )

        assert surface_start_point is not None
        assert surface_end_point is not None

        self.track_length = np.linalg.norm(surface_end_point - surface_start_point)

        if np.all(surface_start_point == surface_end_point):
            raise ValueError("This should not happen")
        if surface_start_point is None:
            raise ValueError("This should not happen")
        if surface_end_point is None:
            raise ValueError("This should not happen")

        return surface_start_point

    def get_angles(self) -> t.Tuple[float, float]:
        """TBW.

        :return:
        """
        beta = np.arccos(
            np.dot(np.array([1.0, 0.0]), np.array([self.dir_ver, self.dir_hor]))
        )

        alpha = np.arccos(
            np.dot(
                np.array([0.0, 0.0, 1.0]),
                np.array([self.dir_ver, self.dir_hor, self.dir_z]),
            )
        )

        if self.dir_hor < 0.0:
            beta += np.pi
            alpha += np.pi

        return alpha, beta

    # def track_length(self):
    #     """TBW.
    #
    #     :return:
    #     """
    #     geo = self.detector.geometry
    #
    #     norm_vectors = [np.array([0., 0., -1.]),    # top plane (usually particle enters vol. via this)
    #                     np.array([0., 0., 1.]),     # bottom plane (usually particle leaves vol. via this)
    #                     np.array([0., 1., 0.]),
    #                     np.array([-1., 0., 0.]),
    #                     np.array([0., -1., 0.]),
    #                     np.array([1., 0., 0.])]
    #
    #     points = [np.array([0., 0., 0.]),                       # top plane (usually particle enters vol. via this)
    #               np.array([0., 0., -1 * geo.total_thickness]), # bottom plane (usually particle leaves vol. via this)
    #               np.array([0., 0., 0.]),
    #               np.array([geo.vert_dimension, 0., 0.]),
    #               np.array([geo.vert_dimension, geo.horz_dimension, 0.]),
    #               np.array([0., geo.horz_dimension, 0.])]
    #
    #     track_length = np.inf
    #     intersect_points = np.zeros((6, 3))
    #     dir_array = np.array([self.dir_ver,
    #                           self.dir_hor,
    #                           self.dir_z])
    #     for i in range(6):
    #         intersect_points[i, :] = find_intersection(n=norm_vectors[i], p0=points[i],
    #                                                    ls=self.starting_position, lv=dir_array)
    #         track_length_new = np.linalg.norm(intersect_points[i, :] - self.starting_position)
    #         if track_length_new < track_length and track_length_new != 0.:
    #             track_length = track_length_new
    #
    #     return track_length

    # TODO: warning this method modify input parameter 'array' !!!
    def intersection_correction(self, array: np.ndarray) -> np.ndarray:
        """TBW.

        :param array:
        :return:
        """
        eps = 1e-8
        geo = self.detector.geometry
        if abs(array[0] - geo.vert_dimension) < eps:
            array[0] = geo.vert_dimension
        if abs(array[0]) < eps:
            array[0] = 0.0
        if abs(array[1] - geo.horz_dimension) < eps:
            array[1] = geo.horz_dimension
        if abs(array[1]) < eps:
            array[1] = 0.0
        if abs(array[2] + geo.total_thickness) < eps:
            array[2] = -1 * geo.total_thickness
        if abs(array[2]) < eps:
            array[2] = 0.0

        return array


def find_intersection(
    n: np.ndarray, p0: np.ndarray, ls: np.ndarray, lv: np.ndarray
) -> np.ndarray:
    """TBW.

    https://en.wikipedia.org/wiki/Line%E2%80%93plane_intersection
    :param n: normal vector of the plane
    :param p0: point of the plane
    :param ls: starting point of particle track
    :param lv: direction of particle track
    :return:
    """
    if np.dot(lv, n) == 0:  # No intersection of track and detector plane
        return None
    else:
        d = np.dot((p0 - ls), n) / np.dot(lv, n)
        p = d * lv + ls
        return p


def isotropic_direction() -> t.Tuple[float, float, float]:
    """TBW.

    :param n:
    :return:
    """
    u = 2 * np.random.random() - 1
    r = np.sqrt(1 - u ** 2)
    kszi = np.random.random()
    v = r * np.cos(2 * np.pi * kszi)
    w = r * np.sin(2 * np.pi * kszi)
    return u, v, w


def non_isotropic_direction(n: int) -> t.Tuple[float, float, float]:
    """TBW.

    :param n:
    :return:
    """
    alpha = 2 * np.pi * np.random.random(n)
    beta = 2 * np.pi * np.random.random(n)
    x = np.cos(alpha) * np.sin(beta)
    y = np.cos(alpha) * np.cos(beta)
    z = np.sin(alpha)
    return x, y, z



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""POPPY (Physical Optics Propagation in Python) model wrapper.

It calculates the optical Point Spread Function of an optical system and applies the convolution.

Documentation:
https://poppy-optics.readthedocs.io/en/stable/index.html

See details about POPPY Optical Element classes:
https://poppy-optics.readthedocs.io/en/stable/available_optics.html

Supported optical elements:

- ``CircularAperture``
- ``SquareAperture``
- ``RectangularAperture``
- ``HexagonAperture``
- ``MultiHexagonalAperture``
- ``ThinLens``
- ``SecondaryObscuration``
- ``ZernikeWFE``
- ``SineWaveWFE``

.. code-block:: yaml

  # YAML config: arguments of POPPY optical_system
  optical_system:
  - item: CircularAperture
    radius: 1.5
  - item: ThinLens
    radius: 1.2
    nwaves: 1
  - item: ZernikeWFE
    radius: 0.8
    coefficients: [0.1e-6, 3.e-6, -3.e-6, 1.e-6, -7.e-7, 0.4e-6, -2.e-6]
    aperture_stop: false
"""

import logging
import typing as t

import matplotlib.pyplot as plt
import numpy as np
import poppy as op
from astropy.convolution import convolve_fft
from astropy.io import fits

from pyxel.data_structure import Photon
from pyxel.detectors import Detector


def calc_psf(
    wavelength: float,
    fov_arcsec: float,
    pixelscale: float,
    optical_system: list,
    display: bool = False,
) -> t.Tuple[t.List[fits.hdu.image.PrimaryHDU], t.List[op.Wavefront]]:
    """Calculate the point spread function for the given optical system and optionally display the psf.

    Parameters
    ----------
    wavelength: float
        Wavelength of incoming light in meters.
    fov_arcsec: float, optional
        Field Of View on detector plane in arcsec.
    pixelscale: float
        Pixel scale on detector plane (arcsec/pixel).
        Defines sampling resolution of PSF.
    optical_system:
        List of optical elements before detector with their specific arguments.

    Returns
    -------
    psf: tuple
        Tuple of lists containing the psf and intermediate wavefronts.
    """
    osys = op.OpticalSystem(npix=1000)  # default: 1024

    for item in optical_system:
        if item["item"] == "CircularAperture":
            optical_item = op.CircularAperture(radius=item["radius"])
        elif item["item"] == "ThinLens":
            optical_item = op.ThinLens(
                nwaves=item["nwaves"],
                reference_wavelength=wavelength,
                radius=item["radius"],
            )
        elif item["item"] == "SquareAperture":
            optical_item = op.SquareAperture(size=item["size"])
        elif item["item"] == "RectangularAperture":
            optical_item = op.RectangleAperture(
                width=item["width"], height=item["height"]
            )  # m
        elif item["item"] == "HexagonAperture":
            optical_item = op.HexagonAperture(side=item["side"])
        elif item["item"] == "MultiHexagonalAperture":
            optical_item = op.MultiHexagonAperture(
                side=item["side"],
                rings=item["rings"],
                gap=item["gap"],
            )  # cm
        elif item["item"] == "SecondaryObscuration":
            optical_item = op.SecondaryObscuration(
                secondary_radius=item["secondary_radius"],
                n_supports=item["n_supports"],
                support_width=item["support_width"],
            )  # cm
        elif item["item"] == "ZernikeWFE":
            optical_item = op.ZernikeWFE(
                radius=item["radius"],
                coefficients=item["coefficients"],  # list of floats
                aperture_stop=item["aperture_stop"],
            )  # bool
        elif item["item"] == "SineWaveWFE":
            optical_item = op.SineWaveWFE(
                spatialfreq=item["spatialfreq"],  # 1/m
                amplitude=item["amplitude"],  # um
                rotation=item["rotation"],
            )
        else:
            raise NotImplementedError
        osys.add_pupil(optical_item)

    osys.add_detector(
        pixelscale=pixelscale,
        fov_arcsec=fov_arcsec,
    )

    psf: t.Tuple[
        t.List[fits.hdu.image.PrimaryHDU], t.List[op.Wavefront]
    ] = osys.calc_psf(
        wavelength=wavelength,
        return_intermediates=True,
        display_intermediates=display,
        normalize="last",
    )

    if display:
        plt.show()

    return psf


def optical_psf(
    detector: Detector,
    wavelength: float,
    fov_arcsec: float,
    pixelscale: float,
    optical_system: list,
) -> None:
    """Model function for poppy optics model: convolve photon array with psf.

    Parameters
    ----------
    detector: Detector
        Pyxel Detector object.
    wavelength: float
        Wavelength of incoming light in meters.
    fov_arcsec: float, optional
        Field Of View on detector plane in arcsec.
    pixelscale: float
        Pixel scale on detector plane (arcsec/pixel).
        Defines sampling resolution of PSF.
    optical_system:
        List of optical elements before detector with their specific arguments.

    Returns
    -------
    None
    """

    logging.getLogger("poppy").setLevel(
        logging.WARNING
    )  # TODO: Fix this. See issue #81

    psf = calc_psf(
        wavelength=wavelength,
        fov_arcsec=fov_arcsec,
        pixelscale=pixelscale,
        optical_system=optical_system,
    )

    # Convolution
    mean = np.mean(detector.photon.array)
    array = convolve_fft(
        detector.photon.array, psf[0][0].data, boundary="fill", fill_value=mean
    )

    detector.photon = Photon(array)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import logging
import operator
import typing as t
from copy import deepcopy
from numbers import Number

import numpy as np

from pyxel.evaluator import eval_entry
from pyxel.pipelines import DetectionPipeline, ModelGroup
from pyxel.state import get_obj_att
from pyxel.util.memory import get_size

if t.TYPE_CHECKING:
    from pyxel.detectors import CCD, CMOS


# TODO: Is this class needed ?
class Processor:
    """TBW."""

    def __init__(self, detector: t.Union["CCD", "CMOS"], pipeline: DetectionPipeline):
        self._log = logging.getLogger(__name__)

        self.detector = detector
        self.pipeline = pipeline

        self._numbytes = 0

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__  # type: str
        return f"{cls_name}<detector={self.detector!r}, pipeline={self.pipeline!r}>"

    def __deepcopy__(self, memodict: dict) -> "Processor":
        return Processor(
            detector=deepcopy(self.detector, memo=memodict),
            pipeline=deepcopy(self.pipeline, memo=memodict),
        )

    # TODO: Could it be renamed '__contains__' ?
    # TODO: reimplement this method.
    def has(self, key: str) -> bool:
        """TBW.

        Parameters
        ----------
        key : str

        Returns
        -------
        bool
            TBW.
        """
        found = False
        obj, att = get_obj_att(self, key)
        if isinstance(obj, dict) and att in obj:
            found = True
        elif hasattr(obj, att):
            found = True
        return found

    # TODO: Could it be renamed '__getitem__' ?
    # TODO: Is it really needed ?
    # TODO: What are the valid keys ? (e.g. 'detector.image.array',
    #       'detector.signal.array' and 'detector.pixel.array')
    def get(self, key: str) -> np.ndarray:
        """TBW.

        Parameters
        ----------
        key : str

        Returns
        -------
        TBW.
        """
        # return get_value(self, key)

        func = operator.attrgetter(key)  # type: t.Callable
        result = func(self)

        return np.asarray(result, dtype=np.float)

    # TODO: Could it be renamed '__setitem__' ?
    def set(
        self,
        key: str,
        value: t.Union[
            str, Number, np.ndarray, t.Sequence[t.Union[str, Number, np.ndarray]]
        ],
        convert_value: bool = True,
    ) -> None:
        """TBW.

        Parameters
        ----------
        key : str
        value
        convert_value : bool
        """
        if convert_value:  # and value:
            # TODO: Refactor this
            # convert the string based value to a number
            if isinstance(value, list):
                for i, val in enumerate(value):
                    if val:
                        value[i] = eval_entry(val)
            else:
                value = eval_entry(value)

        obj, att = get_obj_att(self, key)

        if isinstance(obj, dict) and att in obj:
            obj[att] = value
        else:
            setattr(obj, att, value)

    # TODO: Create a method `DetectionPipeline.run`
    def run_pipeline(self, abort_before: t.Optional[str] = None) -> "Processor":
        """Run a pipeline with all its models in the right order.

        Parameters
        ----------
        abort_before : str
            model name, the pipeline should be aborted before this

        Returns
        -------
        Processor
            TBW.

        Notes
        -----
        The ``Processor`` instance with method '.run_pipeline' is modified.
        """
        self._log.info("Start pipeline")

        # TODO: Use with-statement to set/unset ._is_running
        self.pipeline._is_running = True

        for group_name in self.pipeline.model_group_names:
            # Get a group of models
            models_grp = getattr(
                self.pipeline, group_name
            )  # type: t.Optional[ModelGroup]

            if models_grp:
                self._log.info("Processing group: %r", group_name)

                abort_flag = models_grp.run(
                    detector=self.detector,
                    pipeline=self.pipeline,
                    abort_model=abort_before,
                )
                if abort_flag:
                    break

        self.pipeline._is_running = False

        # TODO: Is is necessary to return 'self' ??
        return self

    @property
    def numbytes(self) -> int:
        """Recursively calculates object size in bytes using Pympler library.

        Returns
        -------
        int
            Size of the object in bytes.
        """
        self._numbytes = get_size(self)
        return self._numbytes



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel photon generator models: photon shot noise."""
import typing as t

import numpy as np

from pyxel.detectors import Detector


# TODO: Fix this
# @validators.validate
# @config.argument(name='seed', label='random seed', units='', validate=checkers.check_type(int))
def shot_noise(detector: Detector, random_seed: t.Optional[int] = None) -> None:
    """Add shot noise to the number of photon per pixel.

    :param detector: Pyxel Detector object
    :param random_seed: int seed
    """
    if random_seed:
        np.random.seed(random_seed)

    detector.photon.array = np.random.poisson(lam=detector.photon.array)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Non-Destructive Readout modes for CMOS-based detectors."""
import logging

from pyxel.detectors import CMOS


# TODO: Fix this
# @validators.validate
# @config.argument(name='mode', label='', units='', validate=check_choices(['uncorrelated', 'CDS', 'Fowler-N', 'UTR']))
# @config.argument(name='fowler_samples', label='', units='', validate=check_type(int))
# @config.argument(name='detector', label='', units='', validate=check_type(CMOS))      # TODO this should be automatic
def non_destructive_readout(detector: CMOS, mode: str, fowler_samples: int = 1) -> None:
    """Non-Destructive Readout modes for CMOS-based detectors.

    :param detector: Pyxel Detector object
    :param mode:
    :param fowler_samples:
    """
    logging.info("")
    if not detector.is_non_destructive_readout or not detector.is_dynamic:
        raise ValueError()

    detector.read_out = False
    if mode == "uncorrelated":
        if detector.time == (detector.end_time - detector.time_step):
            detector.read_out = True
    elif mode == "CDS":
        if detector.time == detector.time_step or detector.time == (
            detector.end_time - detector.time_step
        ):
            detector.read_out = True
    elif mode == "Fowler-N":
        nt = fowler_samples * detector.time_step
        detector.read_out = True
        if nt < detector.time <= detector.end_time - nt:
            detector.read_out = False
    elif mode == "UTR":
        detector.read_out = True



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Simple model to generate charges due to dark current process."""
import logging

import numpy as np

from pyxel.detectors import CMOS


# TODO: Fix this
# @validators.validate
# @config.argument(name='detector', label='', units='', validate=checkers.check_type(CMOS))
def dark_current_rule07(detector: CMOS) -> None:
    """Generate charge from dark current process.

    :param detector: Pyxel Detector object
    TODO: investigate on the knee of rule07 for higher 1/le*T values
    """
    logging.info("")
    geo = detector.geometry
    temperature = detector.environment.temperature
    cutoff = detector.characteristics.cutoff

    amp_to_eps = 6.242e18  # conversion factor from Ampere to Electrons per second
    um2_to_cm2 = 1.0e-8
    conversion_factor = amp_to_eps * um2_to_cm2

    # pitch = 18              # um
    # Rule 07 empirical model parameters
    j0 = 8367.000019  # A/cm**2
    c = -1.162972237
    q = 1.602176624e-19  # Elementary charge (Coulomb)
    k = 1.38064852e-23  # Boltzmann constant (m2 kg s-2 K-1)

    def lambda_e(lambda_cutoff: float) -> float:
        """Compute lambda_e.

        :param lambda_cutoff: (int) Cut-off wavelength of the detector
        """
        lambda_scale = 0.200847413  # um
        lambda_threshold = 4.635136423  # um
        pwr = 0.544071282
        if lambda_cutoff < lambda_threshold:
            le = lambda_cutoff / (
                1
                - ((lambda_scale / lambda_cutoff) - (lambda_scale / lambda_threshold))
                ** pwr
            )
        else:
            le = lambda_cutoff
        return le

    # Rule07
    j = j0 * np.exp(c * (1.24 * q / k) * 1.0 / (lambda_e(cutoff) * temperature))
    dark = j * conversion_factor * geo.pixel_horz_size * geo.pixel_vert_size
    # The number of charge generated with Poisson distribution using rule07 empiric law for lambda
    charge_number = np.random.poisson(dark, size=(geo.row, geo.col))

    charge_number = charge_number.flatten()
    where_non_zero = np.where(charge_number > 0.0)
    charge_number = charge_number[where_non_zero]
    size = charge_number.size

    init_ver_pix_position = geo.vertical_pixel_center_pos_list()[where_non_zero]
    init_hor_pix_position = geo.horizontal_pixel_center_pos_list()[where_non_zero]

    detector.charge.add_charge(
        particle_type="e",
        particles_per_cluster=charge_number,
        init_energy=[0.0] * size,
        init_ver_position=init_ver_pix_position,
        init_hor_position=init_hor_pix_position,
        init_z_position=[0.0] * size,
        init_ver_velocity=[0.0] * size,
        init_hor_velocity=[0.0] * size,
        init_z_velocity=[0.0] * size,
    )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Single outputs."""

import logging
import typing as t
import warnings
from pathlib import Path
from time import strftime

import attr
import h5py as h5
import numpy as np
import pandas as pd
from astropy.io import fits as fits
from matplotlib import pyplot as plt

from pyxel import __version__ as version

from .outputs import PlotArguments, PlotType, apply_run_number, update_plot

if t.TYPE_CHECKING:
    from ..detectors import Detector
    from ..pipelines import Processor

    class SaveToFile(t.Protocol):
        """TBW."""

        def __call__(self, data: np.ndarray, name: str) -> Path:
            """TBW."""
            ...


@attr.s(auto_attribs=True, slots=True, frozen=True)
class SinglePlot:
    """TBW."""

    plot_type: PlotType = attr.ib(converter=PlotType)
    x: str = "detector.photon.array"  # TODO: Check if the value is valid
    y: str = "detector.image.array"  # TODO: Check if the value is valid
    plot_args: t.Optional[PlotArguments] = None


class SingleOutputs:
    """TBW."""

    def __init__(
        self,
        output_folder: t.Union[str, Path],
        save_data_to_file: t.Optional[
            t.Sequence[t.Mapping[str, t.Sequence[str]]]
        ] = None,
        single_plot: t.Optional[SinglePlot] = None,
    ):
        self._log = logging.getLogger(__name__)

        # self.input_file = None  # type: t.Optional[Path]

        # Parameter(s) specific for 'Single'
        self._single_plot = None  # type: t.Optional[SinglePlot]
        if single_plot:
            self._single_plot = single_plot

        self.user_plt_args = None  # type: t.Optional[PlotArguments]
        self.output_dir = (
            Path(output_folder).joinpath("run_" + strftime("%Y%m%d_%H%M%S")).resolve()
        )  # type: Path

        # TODO: Not related to a plot. Use by 'single' and 'parametric' modes.
        self.save_data_to_file = save_data_to_file or [
            {"detector.image.array": ["fits"]}
        ]  # type: t.Sequence[t.Mapping[str, t.Sequence[str]]]

        if self.output_dir.exists():
            raise IsADirectoryError("Directory exists.")

        self.output_dir.mkdir(parents=True, exist_ok=True)

        # self.default_ax_args = {
        #     "xlabel": None,
        #     "ylabel": None,
        #     "title": None,
        #     "axis": None,
        #     "grid": False,
        #     "xscale": "linear",
        #     "yscale": "linear",
        #     "xticks": None,
        #     "yticks": None,
        #     "xlim": [None, None],
        #     "ylim": [None, None],
        #     "sci_x": False,
        #     "sci_y": False,
        # }  # type: dict
        self.default_ax_args = PlotArguments()

        self.default_plot_args = {
            "color": None,
            "marker": ".",
            "linestyle": "",
        }  # type: dict

        self.default_hist_args = {
            "bins": None,
            "range": None,
            "density": None,
            "log": False,
            "cumulative": False,
            "histtype": "step",
            "color": None,
            "facecolor": None,
        }  # type: dict
        self.default_scatter_args = {"size": None, "cbar_label": None}  # type: dict

        self.plt_args = {}  # type: dict

        # TODO: Create an object self._fig here ?
        fig, ax = plt.subplots(1, 1)

        self._fig = fig  # type: plt.Figure
        self._ax = ax  # type: plt.Axes

        plt.close(self._fig)

    def __repr__(self):
        cls_name = self.__class__.__name__  # type: str
        return f"{cls_name}<output_dir={self.output_dir!r}>"

    @property
    def fig(self) -> plt.Figure:
        """Get the current ``Figure``."""
        return self._fig

    def new_file(self, filename: str) -> Path:
        """TBW."""
        new_filename = self.output_dir.joinpath(filename)  # type: Path
        new_filename.touch()

        return new_filename

    # TODO: Specific to 'single_plot' ?
    def save_to_png(self, data: np.ndarray, name: str) -> Path:
        """Write array to bitmap PNG image file.

        Parameters
        ----------
        data : array
        name : str

        Returns
        -------
        Path
            TBW.
        """
        row, col = data.shape
        name = str(name).replace(".", "_")
        filename = apply_run_number(
            self.output_dir.joinpath(f"{name}_??.png")
        )  # type: Path

        dpi = 300
        self._fig.set_size_inches(min(col / dpi, 10.0), min(row / dpi, 10.0))

        # ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0])
        # ax.set_axis_off()
        #
        # fig.add_axes(ax)
        # plt.imshow(data, cmap="gray", extent=[0, col, 0, row])
        self._fig.savefig(filename, dpi=dpi)

        return filename

    def save_to_fits(self, data: np.ndarray, name: str) -> Path:
        """Write array to FITS file.

        Parameters
        ----------
        data
        name

        Returns
        -------
        Path
            TBW.
        """
        name = str(name).replace(".", "_")
        filename = apply_run_number(
            self.output_dir.joinpath(f"{name}_??.fits")
        ).resolve()  # type: Path

        self._log.info("Save to FITS - filename: '%s'", filename)

        hdu = fits.PrimaryHDU(data)
        hdu.header["PYXEL_V"] = (str(version), "Pyxel version")
        hdu.writeto(filename, overwrite=False, output_verify="exception")

        return filename

    def save_to_hdf(self, data: "Detector", name: str) -> Path:
        """Write detector object to HDF5 file."""
        name = str(name).replace(".", "_")
        filename = apply_run_number(self.output_dir.joinpath(f"{name}_??.h5"))
        with h5.File(filename, "w") as h5file:
            h5file.attrs["pyxel-version"] = str(version)
            if name == "detector":
                detector_grp = h5file.create_group("detector")
                for array, name in zip(
                    [
                        data.signal.array,
                        data.image.array,
                        data.photon.array,
                        data.pixel.array,
                        data.charge.frame,
                    ],
                    ["Signal", "Image", "Photon", "Pixel", "Charge"],
                ):
                    dataset = detector_grp.create_dataset(name, np.shape(array))
                    dataset[:] = array
            else:
                detector_grp = h5file.create_group("data")
                dataset = detector_grp.create_dataset(name, np.shape(data))
                dataset[:] = data
        return filename

    def save_to_txt(self, data: np.ndarray, name: str) -> Path:
        """Write data to txt file."""
        name = str(name).replace(".", "_")
        filename = apply_run_number(self.output_dir.joinpath(name + "_??.txt"))
        np.savetxt(filename, data, delimiter=" | ", fmt="%.8e")
        return filename

    def save_to_csv(self, data: pd.DataFrame, name: str) -> Path:
        """Write Pandas Dataframe or Numpy array to a CSV file."""
        name = str(name).replace(".", "_")
        filename = apply_run_number(
            self.output_dir.joinpath(name + "_??.csv")
        )  # type: Path
        try:
            data.to_csv(filename, float_format="%g")
        except AttributeError:
            np.savetxt(filename, data, delimiter=",", fmt="%.8e")
        return filename

    def save_to_npy(self, data: np.ndarray, name: str) -> Path:
        """Write Numpy array to Numpy binary npy file."""
        name = str(name).replace(".", "_")
        filename = apply_run_number(self.output_dir.joinpath(name + "_??.npy"))
        np.save(file=filename, arr=data)
        return filename

    def save_plot(self, filename: str = "figure_??") -> Path:
        """Save plot figure in PNG format, close figure and create new canvas for next plot."""
        new_filename = self.output_dir.joinpath(filename + ".png")  # type: Path
        output_filename = apply_run_number(new_filename).resolve()  # type: Path

        self._log.info("Save plot in filename '%s'", output_filename)
        self._fig.savefig(output_filename)
        # plt.close('all')
        # plt.figure()

        return output_filename

    def plot_graph(
        self, x: np.ndarray, y: np.ndarray, args: t.Optional[dict] = None
    ) -> None:
        """TBW.

        Parameters
        ----------
        x
        y
        args
        """
        ax_args0, plt_args0 = self.update_args(plot_type=PlotType.Graph, new_args=args)

        user_plt_args_dct = None  # type: t.Optional[t.Mapping]
        if isinstance(self.user_plt_args, PlotArguments):
            user_plt_args_dct = self.user_plt_args.to_dict()

        ax_args, plt_args = self.update_args(
            plot_type=PlotType.Graph,
            new_args=user_plt_args_dct,
            ax_args=ax_args0,
            plt_args=plt_args0,
        )

        self._ax.plot(
            x,
            y,
            color=plt_args["color"],
            marker=plt_args["marker"],
            linestyle=plt_args["linestyle"],
        )
        update_plot(ax_args=ax_args, ax=self._ax)
        # plt.draw()

    def plot_histogram(self, data: np.ndarray, args: t.Optional[dict] = None) -> None:
        """TBW.

        Parameters
        ----------
        data
        args
        """
        assert self.user_plt_args is not None

        user_plt_args_dct = None  # type: t.Optional[t.Mapping]
        if isinstance(self.user_plt_args, PlotArguments):
            user_plt_args_dct = self.user_plt_args.to_dict()

        ax_args0, plt_args0 = self.update_args(
            plot_type=PlotType.Histogram, new_args=args
        )
        ax_args, plt_args = self.update_args(
            plot_type=PlotType.Histogram,
            new_args=user_plt_args_dct,
            ax_args=ax_args0,
            plt_args=plt_args0,
        )
        if isinstance(data, np.ndarray):
            data = data.flatten()

        self._ax.hist(
            x=data,
            bins=plt_args["bins"],
            range=plt_args["range"],
            density=plt_args["density"],
            log=plt_args["log"],
            cumulative=plt_args["cumulative"],
            histtype=plt_args["histtype"],
            color=plt_args["color"],
            facecolor=plt_args["facecolor"],
        )

        update_plot(ax_args=ax_args, ax=self._ax)

    def plot_scatter(
        self,
        x: np.ndarray,
        y: np.ndarray,
        color: t.Optional[str] = None,
        args: t.Optional[dict] = None,
    ) -> None:
        """TBW.

        Parameters
        ----------
        x
        y
        color
        args
        """
        user_plt_args_dct = None  # type: t.Optional[t.Mapping]

        if isinstance(self.user_plt_args, PlotArguments):
            user_plt_args_dct = self.user_plt_args.to_dict()

        ax_args0, plt_args0 = self.update_args(
            plot_type=PlotType.Scatter, new_args=args
        )
        ax_args, plt_args = self.update_args(
            plot_type=PlotType.Scatter,
            new_args=user_plt_args_dct,
            ax_args=ax_args0,
            plt_args=plt_args0,
        )

        # fig = plt.gcf()
        # ax = fig.axes

        if color is not None:
            sp = self._ax.scatter(x, y, c=color, s=plt_args["size"])
            cbar = self._fig.colorbar(sp)
            cbar.set_label(plt_args["cbar_label"])

        else:
            self._ax.scatter(x, y, s=plt_args["size"])

        update_plot(ax_args=ax_args, ax=self._ax)
        # plt.draw()
        # fig.canvas.draw_idle()

    # TODO: Specific to 'single_plot'
    # TODO: This function is doing too much.
    def single_output(self, processor: "Processor") -> None:
        """Save data into a file and/or generate pictures.

        Parameters
        ----------
        processor
        """
        warnings.warn(
            "Use function 'save_to_file' and 'single_plot'.", DeprecationWarning
        )

        assert self._single_plot is not None

        # if not self.save_data_to_file:
        #     self.save_data_to_file = [{"detector.image.array": ["fits"]}]

        self.save_to_file(processor)

        self.single_to_plot(processor)

    # TODO: Specific to 'single_plot'
    def single_to_plot(self, processor: "Processor") -> None:
        """Generate picture(s).

        Parameters
        ----------
        processor
        """
        assert self._single_plot is not None

        self.user_plt_args = None
        # todo: default plots with plot_args?
        color = None

        if self._single_plot.plot_args:
            plot_args = self._single_plot.plot_args  # type: PlotArguments
            self.user_plt_args = plot_args

        x = processor.get(self._single_plot.x)
        y = processor.get(self._single_plot.y)

        x = x.flatten()
        y = y.flatten()

        if self._single_plot.plot_type is PlotType.Graph:
            self.plot_graph(x=x, y=y)
            fname = "graph_??"  # type: str

        elif self._single_plot.plot_type is PlotType.Histogram:
            self.plot_histogram(y)
            fname = "histogram_??"

        elif self._single_plot.plot_type is PlotType.Scatter:
            self.plot_scatter(x, y, color)
            fname = "scatter_??"

        else:
            raise NotImplementedError

        self.save_plot(filename=fname)

        # plt.close()

    def save_to_file(self, processor: "Processor") -> t.Sequence[Path]:
        """Save outputs into file(s).

        Parameters
        ----------
        processor : Processor


        Returns
        -------
        list of ``Path``
            TBW.
        """
        save_methods = {
            "fits": self.save_to_fits,
            "hdf": self.save_to_hdf,
            "npy": self.save_to_npy,
            "txt": self.save_to_txt,
            "csv": self.save_to_csv,
            "png": self.save_to_png,
        }  # type: t.Dict[str, SaveToFile]

        filenames = []  # type: t.List[Path]

        for dct in self.save_data_to_file:  # type: t.Mapping[str, t.Sequence[str]]
            # Get first entry of `dict` 'item'
            first_item, *_ = dct.items()

            obj, format_list = first_item
            data = processor.get(obj)  # type: np.ndarray

            if format_list is not None:
                for out_format in format_list:
                    func = save_methods[out_format]  # type: SaveToFile
                    filename = func(data=data, name=obj)  # type: Path

                    filenames.append(filename)

        return filenames

    def update_args(
        self,
        plot_type: PlotType,
        new_args: t.Optional[t.Mapping] = None,
        ax_args: t.Optional[dict] = None,
        plt_args: t.Optional[dict] = None,
    ) -> t.Tuple[dict, dict]:
        """TBW."""
        if new_args is None:
            new_args = {}

        if ax_args is None:
            ax_args = self.default_ax_args.to_dict()

        if plt_args is None:
            if plot_type is PlotType.Histogram:
                plt_args = self.default_hist_args.copy()
            elif plot_type is PlotType.Graph:
                plt_args = self.default_plot_args.copy()
            elif plot_type is PlotType.Scatter:
                plt_args = self.default_scatter_args.copy()
            else:
                raise ValueError

        for key in new_args:
            if key in plt_args.keys():
                plt_args[key] = new_args[key]
            elif key in ax_args.keys():
                ax_args[key] = new_args[key]
            else:
                raise KeyError('Not valid plotting key in "plot_args": "%s"' % key)

        return ax_args, plt_args



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel detector simulation framework."""

# flake8: noqa
try:
    # Try to import library 'attr'
    import attr

    del attr
    WITH_ATTR = True

except ImportError:
    WITH_ATTR = False

from ._version import get_versions
from .show_versions import show_versions

__all__ = ["get_versions", "show_versions"]
__version__ = get_versions()["version"]
del get_versions


if WITH_ATTR:
    # Library 'attr' is installed therefore 'set_options' can be used.
    from .options import SetOptions as set_options

    __all__ += ["set_options"]



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Subpackage containing ``typing.Protocol`` definition."""
import typing as t

import numpy as np
from typing_extensions import Protocol

if t.TYPE_CHECKING:
    import pygmo as pg

__all__ = ["IslandProtocol", "ProblemSingleObjective"]


class IslandProtocol(Protocol):
    """Protocol for a User Define Island."""

    def run_evolve(
        self, algo: "pg.algorithm", pop: "pg.population"
    ) -> t.Tuple["pg.algorithm", "pg.population"]:
        """Run 'evolve' method."""
        ...


class ProblemSingleObjective(Protocol):
    """Create a `Protocol` for a user defined Single Objective `Problem` for Pygmo2.

    A single objective is a deterministic, derivative-free, unconstrained
    optimization problem.

    See https://esa.github.io/pygmo2/problem.html#pygmo.problem.
    """

    def fitness(self, parameter: np.ndarray) -> t.Sequence[float]:
        """Return the fitness of the input decision vector.

        Concatenate the objectives, the equality and the inequality constraints.
        """
        ...

    def get_bounds(self) -> t.Tuple[t.Sequence[float], t.Sequence[float]]:
        """Get the box bounds of the problem (lower_boundary, upper_boundary).

        It also implicitly defines the dimension of the problem.
        """
        ...

    # TODO: Add something about 'batch_fitness'
    #       (see https://esa.github.io/pygmo2/problem.html#pygmo.problem.batch_fitness)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Tools for jupyter notebook visualization."""

import typing as t

import holoviews as hv
import matplotlib.pyplot as plt
import numpy as np

# # Display methods for detector objects in Jupyter notebook
from IPython.display import Markdown, display
from mpl_toolkits.axes_grid1 import make_axes_locatable

if t.TYPE_CHECKING:
    from holoviews import Layout

    # from pyxel.data_structure import Image, Photon, Pixel, Signal
    from pyxel.detectors import Detector
    from pyxel.inputs_outputs import Configuration
    from pyxel.pipelines import DetectionPipeline, ModelFunction, Processor


# ----------------------------------------------------------------------------------------------
# Those two methods are used to display the contents of the configuration once loaded in pyxel


def display_config(configuration: "Configuration", only: str = "all") -> None:
    """Display configuration.

    Parameters
    ----------
    cfg: dict
    only: str

    Returns
    -------
    None
    """
    cfg = configuration.__dict__  # type: dict
    for key in cfg:
        if cfg[key] is None:
            pass
        elif (only not in cfg.keys()) & (only != "all"):
            error = "Config file only contains following keys: " + str(cfg.keys())
            display(Markdown(f"<font color=red> {error} </font>"))
            break
        elif (only == key) & (only != "all"):
            display(Markdown(f"## <font color=blue> {key} </font>"))
            display(Markdown("\t" + str(cfg[key])))
            if isinstance(cfg[key].__dict__, dict):
                display_dict(cfg[key].__dict__)
        elif only == "all":
            display(Markdown(f"## <font color=blue> {key} </font>"))
            display(Markdown("\t" + str(cfg[key])))
            if isinstance(cfg[key].__dict__, dict):
                display_dict(cfg[key].__dict__)


def display_dict(cfg: dict) -> None:
    """Display configuration dictionary.

    Parameters
    ----------
    cfg: dict

    Returns
    -------
    None
    """
    for key in cfg:
        display(Markdown(f"#### <font color=#0088FF> {key} </font>"))
        display(Markdown("\t" + str(cfg[key])))


# ----------------------------------------------------------------------------------------------
# This method will display the parameters of a specific model


def display_model(configuration: "Configuration", model_name: str) -> None:
    """Display model from configuration dictionary or Processor object.

    Parameters
    ----------
    pipeline_container: Processor or dict
    model_name: str

    Returns
    -------
    None
    """

    pipeline = configuration.pipeline  # type: DetectionPipeline
    model = pipeline.get_model(name=model_name)  # type: ModelFunction
    display(Markdown(f"## <font color=blue> {model_name} </font>"))
    display(Markdown(f"Model {model_name} enabled? {model.enabled}"))
    display_dict(model.arguments)


def change_modelparam(
    processor: "Processor", model_name: str, argument: str, changed_value: t.Any
) -> None:
    """Change model parameter.

    Parameters
    ----------
    processor: Processor
    model_name: str
    argument:str
    changed_value

    Returns
    -------
    None
    """
    display(Markdown(f"## <font color=blue> {model_name} </font>"))
    model = processor.pipeline.get_model(name=model_name)  # type: ModelFunction
    model.change_argument(argument, changed_value)
    display(Markdown(f"Changed {argument} to {changed_value}."))


def set_modelstate(processor: "Processor", model_name: str, state: bool = True) -> None:
    """Change model state (true/false).

    Parameters
    ----------
    processor: Processor
    model_name: str
    state: bool

    Returns
    -------
    None
    """
    display(Markdown(f"## <font color=blue> {model_name} </font>"))
    model = processor.pipeline.get_model(name=model_name)  # type: ModelFunction
    model.enabled = state
    display(Markdown(f"Model {model_name} enabled? {model.enabled}"))


# ----------------------------------------------------------------------------------------------
# These method are used to display the detector object (all of the array Photon, pixel, signal and image)


def display_detector(detector: "Detector") -> t.Union["Layout"]:
    """Display detector interactively.

    Parameters
    ----------
    detector: Detector
    hist: bool

    Returns
    -------
    out
        A Holoviews object.
    """
    hv.extension("bokeh")

    # Container for detector data, leave out where there is none.
    det = {}
    if not isinstance(detector.input_image, type(None)):
        det["Input"] = detector.input_image
    if not isinstance(detector._photon, type(None)):
        det["Photon"] = detector.photon.array
    if not isinstance(detector._pixel, type(None)):
        det["Pixel"] = detector.pixel.array
    if not isinstance(detector._signal, type(None)):
        det["Signal"] = detector.signal.array
    if not isinstance(detector._image, type(None)):
        det["Image"] = detector.image.array

    def get_image(name):

        data = det[name]  # type: np.ndarray

        if detector.geometry.row == 1:
            im = hv.Curve((range(len(data[0, :])), data[0, :])).opts(
                tools=["hover"], aspect=1.5, xlabel="x", ylabel="z"
            )
        elif detector.geometry.col == 1:
            im = hv.Curve((range(len(data[:, 0])), data[:, 0])).opts(
                tools=["hover"], aspect=1.5, xlabel="y", ylabel="z"
            )
        else:
            im = hv.Image((range(data.shape[1]), range(data.shape[0]), data)).opts(
                colorbar=True,
                cmap="gray",
                tools=["hover"],
                aspect=(detector.geometry.col / detector.geometry.row),
            )
        return im

    array_names = [key for key in det.keys()]

    if len(array_names) == 0:
        raise ValueError("No data in the detector.")

    dmap = hv.DynamicMap(get_image, kdims=["Array"]).redim.values(Array=array_names)
    dmap = dmap.opts(framewise=True)

    hist = dmap.hist(adjoin=False, num_bins=100).opts(
        aspect=1.5, framewise=True, tools=["hover"], xlabel="z"
    )
    out = (dmap.relabel("Array") + hist.relabel("Histogram")).opts(tabs=True)

    return out


def display_array(data: np.ndarray, axes: t.List[plt.axes], **kwargs: str) -> None:
    """For a pair of axes, display the image on the first one, the histogram on the second.

    Parameters
    ----------
    data: ndarray
        A 2D np.array.
    axes: list
        A list of two axes in a figure.

    Returns
    -------
    None
    """
    im = axes[0].imshow(data)
    axes[0].get_xaxis().set_visible(False)
    axes[0].get_yaxis().set_visible(False)
    divider = make_axes_locatable(axes[0])
    cax1 = divider.append_axes("left", size="5%", pad=0.05)
    axes[0].set_title(kwargs["label"])
    plt.colorbar(im, cax=cax1)
    cax1.yaxis.set_label_position("left")
    cax1.yaxis.set_ticks_position("left")
    axes[1].hist(data.flatten(), bins=50, **kwargs)
    plt.setp(axes[1].xaxis.get_majorticklabels(), rotation=45)
    axes[1].legend(fontsize=12)
    axes[1].grid(True, alpha=0.5)


# def display_detector(
#     detector: "Detector", array: t.Union[None, Photon, Pixel, Signal, Image] = None
# ) -> None:
#     """Display detector.
#
#     Parameters
#     ----------
#     detector: Detector
#     array: str
#
#     Returns
#     -------
#     None
#     """
#     if array is not None:
#         fig, axes = plt.subplots(1, 2, figsize=(15, 6))
#         display_array(array.array, axes, label=str(array).split("<")[0])
#     else:
#         arrays = [detector.photon, detector.pixel, detector.signal, detector.image]
#
#         fig, axes = plt.subplots(len(arrays), 2, figsize=(15, 6 * len(arrays)))
#
#         for idx, data in enumerate(arrays):
#             display_array(data.array, axes[idx], label=str(data).split("<")[0])
#
#     plt.show()


# ----------------------------------------------------------------------------------------------
# These method are used to display the detector memory


def display_persist(persist_dict: dict) -> None:
    """Plot all trapped charges using the memory dict.

    Parameters
    ----------
    persist_dict: dict

    Returns
    -------
    None
    """
    trapped_charges = [trapmap for trapmap in persist_dict.values()]

    fig, axes = plt.subplots(
        len(trapped_charges), 2, figsize=(10, 5 * len(trapped_charges))
    )
    for ax, trapmap, keyw in zip(
        axes, trapped_charges, [i.replace("_", "\n") for i in persist_dict.keys()]
    ):
        display_array(trapmap, ax, label=keyw)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Subpackage containing user defined Island and Batch Fitness evaluator using Dask."""

import logging
import typing as t

import numpy as np
import pygmo as pg
from dask import array as da
from dask import delayed as delayed

__all__ = ["DaskBFE", "DaskIsland"]


class ProblemSerializable:
    """Create a 'problem' with a serializable fitness method.

    Method '.fitness' from a ``pg.problem`` is not serializable with 'cloudpickle'.
    Method ``ProblemSerializable.fitness`` is seriablizable with 'cloudpickle'.

    Examples
    --------
    Create a new pygmo problem and serializable problem
    >>> import pygmo as pg
    >>> prob = pg.problem(...)
    >>> prob_serial = ProblemSerializable(prob)

    Fitness function is working for the pygmo and serializable problem
    >>> import numpy as np
    >>> dvs = np.array([...])
    >>> np.array_equal(prob.fitness(dvs), prob_serial(dvs))
    True

    Serialization is not working for method 'prob.fitness' with 'cloudpickle'
    >>> import cloudpickle
    >>> _ = cloudpickle.dumps(prob.fitness)
    TypeError: cannot pickle 'PyCapsule' object

    Serialization is working with method 'prob_serial.fitness' with 'cloudpickle'
    >>> _ = cloudpickle.dumps(prob_serial.fitness)  # It's working !
    """

    def __init__(self, prob):
        self._prob = prob

    def fitness(self, *args, **kwargs):
        """Compute fitness."""
        return self._prob.fitness(*args, **kwargs)


class AlgoSerializable:
    """Create an 'algorithm' with a serializable evolve method."""

    def __init__(self, algo):
        self._algo = algo

    def evolve(self, *args, **kwargs):
        """Compute 'evolve'."""
        pop = self._algo.evolve(*args, **kwargs)

        return self._algo, pop


class DaskBFE:
    """User defined Batch Fitness Evaluator using `Dask`.

    This class is a user-defined batch fitness evaluator based on 'Dask'.
    """

    def __init__(self, chunk_size: t.Optional[int] = None):
        self._chunk_size = chunk_size

    def __call__(self, prob: "pg.problem", dvs_1d: np.ndarray) -> da.Array:
        """Call operator to run the batch fitness evaluator.

        Parameters
        ----------
        prob
        dvs_1d

        Returns
        -------
        array_like
            A 1d array with the fitness parameters.
        """
        ndims_dvs = prob.get_nx()  # type: int
        num_fitness = prob.get_nf()  # type: int

        if self._chunk_size is None:
            chunk_size = max(1, num_fitness // 10)  # type: int
        else:
            chunk_size = self._chunk_size

        # [dvs_1_1, ..., dvs_1_n, dvs_2_1, ..., dvs_2_n, ..., dvs_m_1, ..., dvs_m_n]

        # [[dvs_1_1, ..., dvs_1_n],
        #  [dvs_2_1, ..., dvs_2_n],
        #  ...
        #  [dvs_m_1, ..., dvs_m_n]]

        # Convert 1D Decision Vectors to 2D `dask.Array`
        dvs_2d = da.from_array(
            dvs_1d.reshape((-1, ndims_dvs)),
            chunks=(chunk_size, ndims_dvs),
        )  # type: da.Array

        logging.info("DaskBFE: %i, %i, %r", len(dvs_1d), ndims_dvs, dvs_2d.shape)

        # Create a new problem with a serializable method '.fitness'
        problem_pickable = ProblemSerializable(prob)

        # Create a generalized function to run a 2D input with 'prob.fitness'
        fitness_func = da.gufunc(
            problem_pickable.fitness,
            signature="(i)->(j)",
            output_dtypes=np.float,
            output_sizes={"j": num_fitness},
            vectorize=True,
        )

        fitness_2d = fitness_func(dvs_2d)  # type: da.Array
        fitness_1d = fitness_2d.ravel()  # type: da.Array

        return fitness_1d

    def get_name(self) -> str:
        """Return name of this evaluator."""
        return "Dask batch fitness evaluator"

    def get_extra_info(self) -> str:
        """Return extra information for this evaluator."""
        return f"Dask batch fitness evaluator with chunk_size={self._chunk_size}."


class DaskIsland:
    """User Defined Island usind `Dask`."""

    def run_evolve(
        self, algo: "pg.algorithm", pop: "pg.population"
    ) -> t.Tuple["pg.algorithm", "pg.population"]:
        """Run 'evolve' method from the input `algorithm` to evolve the input `population`.

        Once the evolution is finished, it will return the algorithm used for the
        evolution and the evolved `population`.

        Parameters
        ----------
        algo : pg.algorithm
            Algorithm used to evolve the input population
        pop : pg.population
            Input population.

        Returns
        -------
        tuple of pg.algorithm, pg.population
            The algorithm used for the evolution and the evolved population.
        """
        logging.info("Run evolve %r, %r", pop, algo)

        # Create a new algorithm with a serializable method '.evolve'
        algo_pickable = AlgoSerializable(algo)

        # Run 'algo.evolve' with `Dask`
        new_delayed_pop = delayed(algo_pickable.evolve, nout=2)(
            pop
        )  # type: delayed.Delayed

        (
            new_algo,
            new_pop,
        ) = new_delayed_pop.compute()  # type: t.Tuple[pg.algo, pg.population]

        return new_algo, new_pop

    def get_name(self) -> str:
        """Return Island's name."""
        return "Dask Island"

    # def get_extra_info(self) -> str:
    #     """Return extra information for this Island."""
    #     pass



"""NGHXRG - Teledyne HxRG Noise Generator.

Modification History:

8-15 April 2015,  B.J. Rauscher,  NASA/GSFC
- Implement Pierre Ferruit's (ESA/ESTEC)recommendation to use
  numpy.fft.rfft and numpy.fft.irfft for faster execution. This saves
  about 30% in execution time.
- Clean up setting default values per several suggestions from
  Chaz Shapiro (NASA/JPL).
- Change from setting the shell variable H2RG_PCA0 to point to the PCA-zero
  file to having a shell variable point to the NG home directory. This is per
  a suggestion from Chaz Shapiro (NASA/JPL) that would allow seamlessly adding
  more PCA-zero templates.
- Implement a request form Chaz Shapiro (NASA/JPL) for status
  reporting. This was done by adding the "verbose" arguement.
- Implement a request from Pierre Ferruit (ESA/ESTEC) to generate
  3-dimensional data cubes.
- Implement a request from Pierre Ferruit to treat ACN as different 1/f
  noise in even/odd columns. Previously ACN was treated purely as a feature
  in Fourier space.
- Version 2(Beta)

16 April 2015,  B.J. Rauscher
- Fixed a bug in the pinkening filter definitions. Abs() was used where
  sqrt() was intended. The bug caused power spectra to have the wrong shape at
  low frequency.
- Version 2.1(Beta)

17 April 2015,  B.J. Rauscher
- Implement a request from Chaz Shapiro for HXRGNoise() to exit gracefully if
  the pca0_file is not found.
- Version 2.2 (Beta)

8 July 2015,  B.J. Rauscher
- Address PASP referee comments
    * Fast scan direction is now reversible. To reverse the slow scan
      direction use the numpy flipud() function.
    * Modifications to support subarrays. Specifically,
        > Setting reference_pixel_border_width=0 (integer zero);
            + (1) eliminates the reference pixel border and
            + (2) turns off adding in a bias pattern
                  when simulating data cubes. (Turned on in v2.5)
- Version 2.4

12 Oct 2015,  J.M. Leisenring,  UA/Steward
- Make compatible with Python 2.x
    * from __future__ import division
- Included options for subarray modes (FULL,  WINDOW,  and STRIPE)
    * Keywords wind_x0 and wind_y0 define a subarray position (lower left corner)
    * Selects correct pca0 region for subarray underlay
    * Adds reference pixel if they exist within subarray window
- Tie negative values to 0 and anything >=2^16 to 2^16-1
- Version 2.5

20 Oct 2015,  J.M. Leisenring,  UA/Steward
- Padded nstep to the next power of 2 in order to improve FFT runtime
    * nstep2 = int(2**np.ceil(np.log2(nstep)))
    * Speeds up FFT calculations by ~5x
- Don't generate noise elements if their magnitudes are equal to 0.
- make_noise() now returns copy of final HDU result for easy retrieval
- Version 2.6

29 Jan 2018,  David Lucsanyi, ESA/ESTEC
- Code has been made PEP8 compatible
- Fixed NGHXRG_HOME file path
- Fixed make_noise() variables
- Bug fixed in ACN mask, now code works with any naxis1, naxis2 and n_out values not just default arguments
- Commented out unused code for ACN arrays
- make_noise() function separated into different noise generator functions,
which returns result numpy array to be able to add to Pyxel detector signal
- removed STRIPE mode (did not work), use WINDOW instead -> BUG: code doesn't work with different x and y array sizes
- Integrated into Pyxel detector simulation framework via a wrapper function
- Version 2.7

20 Feb 2019,  David Lucsanyi, ESA/ESTEC
- use logging
- ZeroDivision error fixed
- Version 2.8
"""
# Necessary for Python 2.6 and later (JML)
# Should still work under Python 3.x (JML)
from __future__ import division, print_function

# warnings.filterwarnings('ignore')
# # Have not verified this in Python 3.x (JML):
# _log = logging.getLogger('nghxrg')
# Have not verified this in Python 3.x (JML):
import logging
import typing as t
from os import path

import numpy as np
from astropy.io import fits
from astropy.stats.funcs import median_absolute_deviation as mad
from scipy.ndimage.interpolation import zoom


def white_noise(nstep: int) -> np.ndarray:
    """Generate white noise for an HxRG including all time steps (actual pixel and overheads).

    Parameters
    ----------
        nstep - Length of vector returned

    """
    return np.random.standard_normal(nstep)


class HXRGNoise:
    """HXRGNoise is a class for making realistic Teledyne HxRG system noise.

    The noise model includes correlated,  uncorrelated,
    stationary,  and non-stationary components. The default parameters
    make noise that resembles Channel 1 of JWST NIRSpec. NIRSpec uses
    H2RG detectors. They are read out using four video outputs at
    1.e+5 pix/s/output.
    """

    # These class variables are common to all HxRG detectors
    NGHXRG_VERSION = 2.8  # Sofware version

    def __init__(
        self,
        det_size_x: int,
        det_size_y: int,
        time_step: int,
        n_out: int,
        nroh: int,
        nfoh: int,
        reverse_scan_direction: bool,
        reference_pixel_border_width: int,
        wind_mode: str,
        wind_x_size: int,
        wind_y_size: int,
        wind_x0: int,
        wind_y0: int,
        verbose: bool,
        pca0_file: t.Optional[str] = None,
    ):
        """Simulate Teledyne HxRG+SIDECAR ASIC system noise.

        Parameters
        ----------
            naxis1      - X-dimension of the FITS cube
            naxis2      - Y-dimension of the FITS cube
            time_step   - Z-dimension of the FITS cube, number of frame in series over time
                          (number of up-the-ramp samples)
            n_out       - Number of detector outputs
            nfoh        - New frame overhead in rows. This allows for a short
                          wait at the end of a frame before starting the next
                          one.
            nroh        - New row overhead in pixel. This allows for a short
                          wait at the end of a row before starting the next one.
            pca0_file   - Name of a FITS file that contains PCA-zero
            verbose     - Enable this to provide status reporting
            wind_mode   - 'FULL' or 'WINDOW' (JML)
            wind_x0/wind_y0       - Pixel positions of subarray mode (JML)
            det_size_x    - Pixel dimension of full detector (square),  used only
                          for WINDOW mode (JML)
            reference_pixel_border_width - Width of reference pixel border
                                           around image area
            reverse_scan_direction - Enable this to reverse the fast scanner
                                     readout directions. This
                                     capability was added to support
                                     Teledyne's programmable fast scan
                                     readout directions. The default
                                     setting =False corresponds to
                                     what HxRG detectors default to
                                     upon power up.

        :param det_size_x:
        :param det_size_y:
        :param time_step:
        :param n_out:
        :param nroh:
        :param nfoh:
        :param pca0_file:
        :param reverse_scan_direction:
        :param reference_pixel_border_width:
        :param wind_mode:
        :param wind_x_size:
        :param wind_y_size:
        :param wind_x0:
        :param wind_y0:
        :param verbose:

        """
        # ======================================================================
        #
        # DEFAULT CLOCKING PARAMETERS
        #
        # The following parameters define the default HxRG clocking pattern. The
        # parameters that define the default noise model are defined in the
        # make_noise() method.
        #
        # ======================================================================
        self._log = logging.getLogger("nghxrg")

        wind_mode = wind_mode.upper()
        if wind_mode == "WINDOW":
            n_out = 1
            self.naxis1 = wind_x_size
            self.naxis2 = wind_y_size
        elif wind_mode == "FULL":
            wind_x0 = 0
            wind_y0 = 0
            self.naxis1 = det_size_x
            self.naxis2 = det_size_y
        # elif wind_mode == 'STRIPE':
        #     wind_x0 = 0
        else:
            raise ValueError("Not a valid window readout mode!")
            # _log.warning('%s not a valid window readout mode! Returning...' % inst_params['wind_mode'])  # ERROR WTF
            # os.sys.exit()

        self.naxis3 = 1  # always use only one frame (2d array)
        self.time_step = time_step  # number of frame over time
        self.n_out = n_out
        # self.dt = dt
        self.nroh = nroh
        self.nfoh = nfoh
        self.reference_pixel_border_width = reference_pixel_border_width
        # self.pca0_file = pca0_file

        # Default clocking pattern is JWST NIRSpec
        # if naxis1 is None:
        #     self.naxis1 = 2048
        # if naxis2 is None:
        #     self.naxis2 = 2048
        # if naxis3 is None:
        #     self.naxis3 = 1
        # if n_out is None:
        #     self.n_out = 4
        # if dt is None:
        #     self.dt = 1.e-5
        # if nroh is None:
        #     self.nroh = 12
        # if nfoh is None:
        #     self.nfoh = 1
        # if reference_pixel_border_width is None:
        #     self.reference_pixel_border_width = 4

        # Check that det_size_x is greater than self.naxis1 and self.naxis2 in WINDOW mode (JML)
        if wind_mode == "WINDOW":
            if self.naxis1 > det_size_x:
                raise ValueError()
                # _log.warning('NAXIS1 %s greater than det_size_x %s! Returning...' % (self.naxis1, det_size_x))
                # os.sys.exit()
            if self.naxis2 > det_size_y:
                raise ValueError()
                # _log.warning('NAXIS2 %s greater than det_size_y %s! Returning...' % (self.naxis1, det_size_y))
                # os.sys.exit()

        # ======================================================================

        # Configure Subarray (JML)
        self.wind_mode = wind_mode
        self.det_size_x = det_size_x
        self.det_size_y = det_size_y
        self.wind_x0 = wind_x0
        self.wind_y0 = wind_y0

        # Select region of pca0 associated with window position
        if self.wind_mode == "WINDOW":
            x1 = self.wind_x0
            y1 = self.wind_y0
        # elif self.wind_mode == 'STRIPE':
        #     x1 = 0
        #     y1 = self.wind_y0
        else:
            x1 = 0
            y1 = 0

        # print(y1, self.naxis2) This appears to be a stub
        x2 = x1 + self.naxis1
        y2 = y1 + self.naxis2

        # How many reference pixel on each border?
        w = self.reference_pixel_border_width  # Easier to work with
        lower = w - y1
        upper = w - (self.det_size_y - y2)
        left = w - x1
        right = w - (self.det_size_x - x2)
        ref_all = np.array([lower, upper, left, right])
        ref_all[ref_all < 0] = 0
        self.ref_all = ref_all

        # Configure status reporting
        self.verbose = verbose

        # Configure readout direction
        self.reverse_scan_direction = reverse_scan_direction

        # Compute the number of pixel in the fast-scan direction per output
        self.xsize = self.naxis1 // self.n_out

        # Compute the number of time steps per integration,  per output
        self.nstep = (
            (self.xsize + self.nroh) * (self.naxis2 + self.nfoh) * self.time_step
        )
        # Pad nsteps to a power of 2,  which is much faster (JML)
        self.nstep2 = int(2 ** np.ceil(np.log2(self.nstep)))

        # For adding in ACN, it is handy to have masks of the even
        # and odd pixel on one output neglecting any gaps
        # UNUSED CODE COMMENTED OUT
        # self.m_even = np.zeros((self.naxis3, self.naxis2, self.xsize))
        # self.m_odd = np.zeros_like(self.m_even)
        # for x in np.arange(0, self.xsize, 2):
        #     self.m_even[:, :self.naxis2, x] = 1
        #     self.m_odd[:, :self.naxis2, x+1] = 1
        # self.m_even = np.reshape(self.m_even,  np.size(self.m_even))
        # self.m_odd = np.reshape(self.m_odd,  np.size(self.m_odd))
        # UNUSED CODE COMMENTED OUT

        # Also for adding in ACN,  we need a mask that point to just
        # the real pixel in ordered vectors of just the even or odd pixel - BUG FIXED HERE (D.L.)
        self.m_short = np.zeros(
            (self.naxis3, (self.naxis2 + self.nfoh) // 2, self.xsize + self.nroh)
        )
        self.m_short[:, : self.naxis2 // 2, : self.xsize] = 1
        # self.m_short = np.reshape(self.m_short,  np.size(self.m_short))
        self.m_short = self.m_short.flatten()

        # Define frequency arrays
        self.f1 = np.fft.rfftfreq(self.nstep2)  # Frequencies for nstep elements
        self.f2 = np.fft.rfftfreq(2 * self.nstep2)  # ... for 2*nstep elements

        # Define pinkening filters. F1 and p_filter1 are used to
        # generate ACN. F2 and p_filter2 are used to generate 1/f noise.
        self.alpha = -1  # Hard code for 1/f noise until proven otherwise
        self.p_filter1 = np.sqrt(self.f1[1:] ** self.alpha)
        self.p_filter2 = np.sqrt(self.f2[1:] ** self.alpha)
        self.p_filter1 = np.insert(self.p_filter1, 0, 0.0)
        self.p_filter2 = np.insert(self.p_filter2, 0, 0.0)

        if pca0_file is not None:
            # Initialize PCA-zero file and make sure that it exists and is a file
            # self.pca0_file = os.getenv('NGHXRG_HOME')+'/nirspec_pca0.fits' if \
            # if pca0_file is None:
            #     self.pca0_file = NGHXRG_HOME + '/nirspec_pca0.fits'

            if path.isfile(pca0_file) is False:
                print("There was an error finding pca0_file! Check to be")
                print("sure that the NGHXRG_HOME shell environment")
                print("variable is set correctly and that the")
                print("$NGHXRG_HOME/ directory contains the desired PCA0")
                print("file. The default is nirspec_pca0.fits.")
                raise ValueError()
                # os.sys.exit()

            # Initialize pca0. This includes scaling to the correct size,
            # zero offsetting,  and renormalization. We use robust statistics
            # because pca0 is real data
            hdu = fits.open(pca0_file)
            nx_pca0 = hdu[0].header["naxis1"]
            ny_pca0 = hdu[0].header["naxis2"]

            zoom_factor = 0
            # Do this slightly differently,  taking into account the
            # different types of readout modes (JML)
            # if (nx_pca0 != self.naxis1 or naxis2 != self.naxis2):
            #    zoom_factor = self.naxis1 / nx_pca0
            #    self.pca0 = zoom(hdu[0].data,  zoom_factor,  order=1,  mode='wrap')
            # else:
            #    self.pca0 = hdu[0].data
            # self.pca0 -= np.median(self.pca0) # Zero offset
            # self.pca0 /= (1.4826*mad(self.pca0)) # Renormalize

            data = hdu[0].data
            # Make sure the real PCA image is correctly scaled to size of fake data (JML)
            # Depends if we're FULL, STRIPE or WINDOW
            if wind_mode == "FULL":
                scale1 = self.naxis1 / nx_pca0
                scale2 = self.naxis2 / ny_pca0
                zoom_factor = np.max([scale1, scale2])
            # if wind_mode == 'STRIPE':
            #     zoom_factor = self.naxis1 / nx_pca0
            if wind_mode == "WINDOW":
                # Scale based on det_size_x
                scale1 = self.det_size_x / nx_pca0
                scale2 = self.det_size_y / ny_pca0
                zoom_factor = np.max([scale1, scale2])

            # Resize PCA0 data
            if zoom_factor != 1:
                data = zoom(data, zoom_factor, order=1, mode="wrap")

            data -= np.median(data)  # Zero offset
            data /= 1.4826 * mad(data)  # Renormalize

            # Make sure x2 and y2 are valid
            if x2 > data.shape[0] or y2 > data.shape[1]:
                raise ValueError()
                # _log.warning('Specified window size does not fit within detector array!')
                # _log.warning('X indices: [%s, %s]; Y indices: [%s, %s]; XY Size: [%s,  %s]' %
                #              (x1, x2, y1, y2, data.shape[0], data.shape[1]))
                # os.sys.exit()
            self.pca0 = data[y1:y2, x1:x2]

    # TODO: Remove this method, use only 'self._log.info'
    def message(self, message_text: str) -> None:
        """Print a message to the terminal."""
        if self.verbose is True:
            self._log.info("NG: " + message_text)

    def pink_noise(self, mode: str) -> np.ndarray:
        """TBW.

        Generate a vector of non-periodic pink noise.

        Parameters
        ----------
            mode - Selected from {'pink',  'acn'}

        """
        # Configure depending on mode setting
        if mode == "pink":
            nstep = 2 * self.nstep
            nstep2 = 2 * self.nstep2  # JML
            # f = self.f2
            p_filter = self.p_filter2
        else:
            nstep = self.nstep
            nstep2 = self.nstep2  # JML
            # f = self.f1
            p_filter = self.p_filter1

        # Generate seed noise
        mynoise = white_noise(nstep2)

        # Save the mean and standard deviation of the first
        # half. These are restored later. We do not subtract the mean
        # here. This happens when we multiply the FFT by the pinkening
        # filter which has no power at f=0.
        the_mean = np.mean(mynoise[: nstep2 // 2])
        the_std = np.std(mynoise[: nstep2 // 2])

        # Apply the pinkening filter.
        thefft = np.fft.rfft(mynoise)
        thefft = np.multiply(thefft, p_filter)
        result = np.fft.irfft(thefft)
        result = result[: nstep // 2]  # Keep 1st half of nstep

        # Restore the mean and standard deviation
        result *= the_std / np.std(result)
        result = result - np.mean(result) + the_mean

        # if mode == 'acn':

        # Done
        return result

    # def make_noise(self, rd_noise=5.2, c_pink=3, u_pink=1, acn=0.5, pca0_amp=0.2,
    #                reference_pixel_noise_ratio=0.8, ktc_noise=29., bias_offset=5000., bias_amp=500.):
    #     """
    #     Generate a FITS cube containing only noise.
    #
    #     Parameters:
    #         rd_noise - Standard deviation of read noise in electrons
    #         c_pink   - Standard deviation of correlated pink noise in electrons
    #         u_pink   - Standard deviation of uncorrelated pink noise in
    #                    electrons
    #         acn      - Standard deviation of alternating column noise in
    #                    electrons
    #         pca0_amp - Standard deviation of pca0 in electrons
    #         reference_pixel_noise_ratio - Ratio of the standard deviation of
    #                                       the reference pixel to the regular
    #                                       pixel. Reference pixel are usually
    #                                       a little lower noise.
    #         ktc_noise   - kTC noise in electrons. Set this equal to
    #                       sqrt(k*T*C_pixel)/q_e,  where k is Boltzmann's
    #                       constant,  T is detector temperature,  and C_pixel is
    #                       pixel capacitance. For an H2RG,  the pixel capacitance
    #                       is typically about 40 fF.
    #         bias_offset - On average,  integrations start here in electrons. Set
    #                       this so that all pixel are in range.
    #         bias_amp    - A multiplicative factor that we multiply PCA-zero by
    #                       to simulate a bias pattern. This is completely
    #                       independent from adding in "picture frame" noise.
    #
    #         pedestal - NOT IMPLEMENTED! Magnitude of pedestal drift in electrons
    #
    #     Note1:
    #     Because of the noise correlations,  there is no simple way to
    #     predict the noise of the simulated images. However,  to a
    #     crude first approximation,  these components add in
    #     quadrature.
    #
    #     Note2:
    #     The units in the above are mostly "electrons". This follows convention
    #     in the astronomical community. From a physics perspective,  holes are
    #     actually the physical entity that is collected in Teledyne's p-on-n
    #     (p-type implants in n-type bulk) HgCdTe architecture.
    #
    #     :param rd_noise:
    #     :param c_pink:
    #     :param u_pink:
    #     :param acn:
    #     :param pca0_amp:
    #     :param reference_pixel_noise_ratio:
    #     :param ktc_noise:
    #     :param bias_offset:
    #     :param bias_amp:
    #     :return:
    #     """
    #
    #     self.message('Starting make_noise()')

    # ======================================================================
    #
    # DEFAULT NOISE PARAMETERS
    #
    # These defaults create noise similar to that seen in the JWST NIRSpec.
    #
    # ======================================================================
    # if rd_noise is None:
    #     rd_noise = 5.2
    # # if pedestal is None:
    # #     pedestal = 4
    # if c_pink is None:
    #     c_pink = 3
    # if u_pink is None:
    #     u_pink = 1
    # if acn is None:
    #     acn = 0.5
    # if pca0_amp is None:
    #     pca0_amp = 0.2

    # Change this only if you know that your detector is different from a
    # typical H2RG.
    # if reference_pixel_noise_ratio is None:
    #     reference_pixel_noise_ratio = 0.8

    # These are used only when generating cubes. They are
    # completely removed when the data are calibrated to
    # correlated double sampling or slope images. We include
    # them in here to make more realistic looking raw cubes.
    # if ktc_noise is None:
    #     ktc_noise = 29.
    # if bias_offset is None:
    #     bias_offset = 5000.
    # if bias_amp is None:
    #     bias_amp = 500.

    # ======================================================================

    # Initialize the result cube. For up-the-ramp integrations,
    # we also add a bias pattern. Otherwise,  we assume
    # that the aim was to simulate a two dimensional correlated
    # double sampling image or slope image.
    # self.message('Initializing results cube')
    # result = np.zeros((self.naxis3,  self.naxis2,  self.naxis1), dtype=np.float32)

    def add_ktc_bias_noise(
        self,
        ktc_noise: float = 29.0,
        bias_offset: float = 5000.0,
        bias_amp: float = 500.0,
    ) -> np.ndarray:
        """TBW.

        Inject a bias pattern and kTC noise.
        :param ktc_noise:
        :param bias_amp:
        :param bias_offset:
        :return:
        """
        result = np.zeros((self.naxis3, self.naxis2, self.naxis1), dtype=np.float32)

        if (
            self.time_step > 1
        ):  # NOTE: there is no kTc or Bias noise added for first/single frame
            self.message("Generating ktc_bias_noise")
            # If there are no reference pixel,
            # we know that we are dealing with a subarray. In this case,  we do not
            # inject any bias pattern for now.
            # if self.reference_pixel_border_width > 0:
            #    bias_pattern = self.pca0*bias_amp + bias_offset
            # else:
            #    bias_pattern = bias_offset

            # Always inject bias pattern. Works for WINDOW and STRIPE (JML)
            bias_pattern = self.pca0 * bias_amp + bias_offset

            # Add in some kTC noise. Since this should always come out
            # in calibration,  we do not attempt to model it in detail.
            bias_pattern += ktc_noise * np.random.standard_normal(
                (self.naxis2, self.naxis1)
            )

            # Ensure that there are no negative pixel values. Data cubes
            # are converted to unsigned integer before writing.
            # bias_pattern = np.where(bias_pattern < 0,  0,  bias_pattern)
            # Updated to conform to Python >=2.6. (JML)
            # bias_pattern[bias_pattern < 0] = 0
            # Actually,  I think this makes the most sense to do at the very end (JML)

            # Add in the bias pattern
            for z in np.arange(self.naxis3):
                result[z, :, :] += bias_pattern

        return result

    def add_white_read_noise(
        self, rd_noise: float = 5.2, reference_pixel_noise_ratio: float = 0.8
    ) -> np.ndarray:
        """TBW.

        Make white read noise. This is the same for all pixel.

        :param rd_noise:
        :param reference_pixel_noise_ratio:
        :return:
        """
        result = np.zeros((self.naxis3, self.naxis2, self.naxis1), dtype=np.float32)

        if rd_noise > 0:
            self.message("Generating rd_noise")
            w = self.ref_all
            r = reference_pixel_noise_ratio  # Easier to work with
            for z in np.arange(self.naxis3):
                here = np.zeros((self.naxis2, self.naxis1))

                # Noisy reference pixel for each side of detector
                if w[0] > 0:  # lower
                    here[: w[0], :] = (  # noqa: E203
                        r * rd_noise * np.random.standard_normal((w[0], self.naxis1))
                    )
                if w[1] > 0:  # upper
                    here[-w[1] :, :] = (  # noqa: E203
                        r * rd_noise * np.random.standard_normal((w[1], self.naxis1))
                    )
                if w[2] > 0:  # left
                    here[:, : w[2]] = (
                        r * rd_noise * np.random.standard_normal((self.naxis2, w[2]))
                    )
                if w[3] > 0:  # right
                    here[:, -w[3] :] = (  # noqa: E203
                        r * rd_noise * np.random.standard_normal((self.naxis2, w[3]))
                    )

                # Noisy regular pixel
                if np.sum(w) > 0:  # Ref. pixel exist in frame
                    start_y_idx = w[0]
                    end_y_idx = self.naxis2 - w[1]

                    start_x_idx = w[2]
                    end_x_idx = self.naxis1 - w[3]

                    here[
                        start_y_idx:end_y_idx, start_x_idx:end_x_idx
                    ] = rd_noise * np.random.standard_normal(
                        (self.naxis2 - w[0] - w[1], self.naxis1 - w[2] - w[3])
                    )
                else:  # No Ref. pixel,  so add only regular pixel
                    here = rd_noise * np.random.standard_normal(
                        (self.naxis2, self.naxis1)
                    )

                # Add the noise in to the result
                result[z, :, :] += here

        return result

    def add_corr_pink_noise(self, c_pink: float = 3.0) -> np.ndarray:
        """TBW.

        Add correlated pink noise.
        :param c_pink:
        :return:
        """
        result = np.zeros((self.naxis3, self.naxis2, self.naxis1), dtype=np.float32)

        if c_pink > 0:
            self.message("Adding c_pink noise")
            tt = c_pink * self.pink_noise("pink")  # tt is a temp. variable
            tt = np.reshape(
                tt, (self.time_step, self.naxis2 + self.nfoh, self.xsize + self.nroh)
            )[:, : self.naxis2, : self.xsize]
            tt = tt[-1, :, :]
            for op in np.arange(self.n_out):
                wind_x0 = op * self.xsize
                x1 = wind_x0 + self.xsize
                # By default fast-scan readout direction is [-->, <--, -->, <--]
                # If reverse_scan_direction is True,  then [<--, -->, <--, -->]
                # Would be nice to include option for all --> or all <--
                if self.reverse_scan_direction:
                    mod_num = 1
                else:
                    mod_num = 0
                if np.mod(op, 2) == mod_num:
                    result[:, :, wind_x0:x1] += tt
                else:
                    result[:, :, wind_x0:x1] += tt[:, :, ::-1]

        return result

    def add_uncorr_pink_noise(self, u_pink: float = 1.0) -> np.ndarray:
        """TBW.

        Add uncorrelated pink noise. Because this pink noise is stationary and
        different for each output,  we don't need to flip it.
        :param u_pink:
        :return:
        """
        result = np.zeros((self.naxis3, self.naxis2, self.naxis1), dtype=np.float32)

        if u_pink > 0:
            self.message("Adding u_pink noise")
            for op in np.arange(self.n_out):
                wind_x0 = op * self.xsize
                x1 = wind_x0 + self.xsize
                tt = u_pink * self.pink_noise("pink")
                tt = np.reshape(
                    tt,
                    (self.time_step, self.naxis2 + self.nfoh, self.xsize + self.nroh),
                )[:, : self.naxis2, : self.xsize]
                tt = tt[-1, :, :]
                result[:, :, wind_x0:x1] += tt

        return result

    def add_acn_noise(self, acn: float = 0.5) -> np.ndarray:
        """TBW.

        Add Alternating Column Noise (ACN)
        :param acn:
        :return:
        """
        result = np.zeros((self.naxis3, self.naxis2, self.naxis1), dtype=np.float32)

        if acn > 0:
            self.message("Adding acn noise")
            for op in np.arange(self.n_out):
                # Generate new pink noise for each even and odd vector.
                # We give these the abstract names 'a' and 'b' so that we
                # can use a previously worked out formula to turn them
                # back into an image section.
                a = acn * self.pink_noise("acn")
                b = acn * self.pink_noise("acn")

                # Pick out just the real pixel (i.e. ignore the gaps)
                a = a[np.where(self.m_short == 1)]
                b = b[np.where(self.m_short == 1)]

                half_ch_pixels = self.naxis1 * self.naxis2 // (2 * self.n_out)
                if len(a) != half_ch_pixels:
                    ValueError(
                        "This should not happen: in ACN noise len(a) != number of half ch pixel"
                    )
                    # a = np.append(a, np.zeros(halfchpixels-len(a)))
                if len(b) != half_ch_pixels:
                    ValueError(
                        "This should not happen: in ACN noise len(b) != number of half ch pixel"
                    )
                    # b = np.append(b, np.zeros(halfchpixels-len(b)))

                # Reformat into an image section. This uses the formula mentioned above.
                acn_cube = np.reshape(
                    np.transpose(np.vstack((a, b))),
                    (self.naxis3, self.naxis2, self.xsize),
                )

                # Add in the ACN. Because pink noise is stationary,  we can
                # ignore the readout directions. There is no need to flip
                # acn_cube before adding it in.
                wind_x0 = op * self.xsize
                x1 = wind_x0 + self.xsize
                result[:, :, wind_x0:x1] += acn_cube

        return result

    # TODO add pca0_file
    def add_pca_zero_noise(self, pca0_amp: float = 0.2) -> np.ndarray:
        """TBW.

        Add PCA-zero. The PCA-zero template is modulated by 1/f.

        :param pca0_amp:
        :return:
        """
        result = np.zeros((self.naxis3, self.naxis2, self.naxis1), dtype=np.float32)

        if pca0_amp > 0:
            self.message('Adding PCA-zero "picture frame" noise')
            gamma = self.pink_noise(mode="pink")
            zoom_factor = self.naxis2 * self.time_step / np.size(gamma)
            gamma = zoom(gamma, zoom_factor, order=1, mode="mirror")
            gamma = np.reshape(gamma, (self.time_step, self.naxis2))
            gamma = gamma[-1, :]
            # for z in np.arange(self.naxis3):
            for y in np.arange(self.naxis2):
                result[0, y, :] += pca0_amp * self.pca0[y, :] * gamma[y]

        return result

    # TODO: Warning this method modify input parameter 'result' !!
    def format_result(self, result: np.ndarray) -> np.ndarray:
        """TBW.

        If the data cube has only 1 frame,  reformat into a 2-dimensional image.

        :param result:
        :return:
        """
        if self.naxis3 == 1:
            self.message("Reformatting cube into image")
            result = result[0, :, :]

        # If the data cube has more than one frame,  convert to unsigned
        # integer
        if self.naxis3 > 1:
            # Data will be converted to 16-bit unsigned int
            # Ensure that there are no negative pixel values.
            result[result < 0] = 0
            # And that anything higher than 65535 gets tacked to the top end
            result[result >= 2 ** 16] = 2 ** 16 - 1

            # self.message('Converting to 16-bit unsigned integer')
            # result = result.astype('uint16')
            result = result.astype("float64")

        self.message("Exiting make_noise()")

        return result

    def create_hdu(self, result: np.ndarray, o_file: t.Optional[str] = None) -> None:
        """TBW.

        Create HDU file and saving data to it
        :return:
        """
        hdu = fits.PrimaryHDU(result)
        # hdu.header.append()
        # hdu.header.append(('RD_NOISE',  rd_noise,  'Read noise'))
        # # hdu.header.append(('PEDESTAL',  pedestal,  'Pedestal drifts'))
        # hdu.header.append(('C_PINK',  c_pink,  'Correlated pink'))
        # hdu.header.append(('U_PINK',  u_pink,  'Uncorrelated pink'))
        # hdu.header.append(('ACN',  acn,  'Alternating column noise'))
        # hdu.header.append(('PCA0',  pca0_amp, 'PCA zero,  AKA picture frame'))
        # hdu.header['HISTORY'] = 'Created by NGHXRG version ' + str(self.nghxrg_version)

        # Write the result to a FITS file
        if o_file is not None:
            self.message("Writing FITS file")
            hdu.writeto(o_file, clobber="True")



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel photon generator models."""
import typing as t

import numpy as np

from pyxel.data_structure import Photon
from pyxel.detectors import Detector


# TODO: Fix this
# @validators.validate
# @config.argument(name='level', label='number of photon', units='', validate=check_type(int))
# @config.argument(name='option', label='type of illumination', units='',
#                  validate=check_choices(['uniform', 'rectangular_hole', 'elliptic_hole']))
# @config.argument(name='size', label='size of 2d array', units='', validate=check_type(list))
# @config.argument(name='hole_size', label='size of hole', units='', validate=check_type(list))
def illumination(
    detector: Detector,
    level: int,
    option: str = "uniform",
    array_size: t.Optional[t.Tuple[int, int]] = None,
    hole_size: t.Optional[t.Sequence[int]] = None,
    hole_center: t.Optional[t.Sequence[int]] = None,
) -> None:
    """Generate photon uniformly over the entire array or hole.

    detector: Detector
        Pyxel Detector object.
    level: int
        Number of photon per pixel.
    option: str{'uniform', 'elliptic_hole', 'rectangular_hole'}
        A string indicating the type of illumination:

        - ``uniform``
           Uniformly fill the entire array with photon. (Default)
        - ``elliptic_hole``
           Mask with elliptic hole.
        - ``rectangular_hole``
           Mask with rectangular hole.
    array_size: list, optional
        List of integers defining the size of 2d photon array.
    hole_size: list, optional
        List of integers defining the sizes of the elliptic or rectangular hole.
    hole_center: list, optional
        List of integers defining the center of the elliptic or rectangular hole.
    """
    if array_size is None:
        if not detector.has_photon:
            geo = detector.geometry
            detector.photon = Photon(np.zeros((geo.row, geo.col), dtype=int))

        shape = detector.photon.array.shape  # type: t.Tuple[int, int]
    else:
        shape = array_size

    if option == "uniform":
        photon_array = np.ones(shape, dtype=int) * level

    elif option == "rectangular_hole":
        if hole_size:
            photon_array = np.zeros(shape, dtype=int)
            if hole_center is not None:
                if not (
                    (0 <= hole_center[0] <= shape[0])
                    and (0 <= hole_center[1] <= shape[1])
                ):
                    raise ValueError(
                        'Argument "hole_center" should be inside Photon array.'
                    )
            else:
                hole_center = [int(shape[0] / 2), int(shape[1] / 2)]
            p = hole_center[0] - int(hole_size[0] / 2)
            q = hole_center[1] - int(hole_size[1] / 2)
            p0 = int(np.clip(p, a_min=0, a_max=shape[0]))
            q0 = int(np.clip(q, a_min=0, a_max=shape[1]))
            photon_array[
                slice(p0, p + hole_size[0]), slice(q0, q + hole_size[1])
            ] = level
        else:
            raise ValueError(
                "hole_size argument should be defined for illumination model"
            )

    elif option == "elliptic_hole":
        if hole_size:
            photon_array = np.zeros(shape, dtype=int)
            if hole_center is not None:
                if not (
                    (0 <= hole_center[0] <= shape[0])
                    and (0 <= hole_center[1] <= shape[1])
                ):
                    raise ValueError(
                        'Argument "hole_center" should be inside Photon array.'
                    )
            else:
                hole_center = [int(shape[0] / 2), int(shape[1] / 2)]
            y, x = np.ogrid[: shape[0], : shape[1]]
            dist_from_center = np.sqrt(
                ((x - hole_center[1]) / hole_size[1]) ** 2
                + ((y - hole_center[0]) / hole_size[0]) ** 2
            )
            photon_array[dist_from_center < 1] = level
        else:
            raise ValueError(
                "hole_size argument should be defined for illumination model"
            )

    else:
        raise NotImplementedError

    try:
        detector.photon.array += photon_array
    except TypeError:
        detector.photon = Photon(photon_array)
    except ValueError:
        raise ValueError("Shapes of arrays do not match")



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""TBW."""

import typing as t

if t.TYPE_CHECKING:
    from ..inputs_outputs import SingleOutputs


class Single:
    """TBW."""

    def __init__(self, outputs: "SingleOutputs"):
        self.outputs = outputs

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__  # type: str
        return f"{cls_name}<outputs={self.outputs!r}>"



"""Unittests for the 'ModelFitting' class."""
#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from pathlib import Path

import numpy as np
import pytest

from pyxel import inputs_outputs as io
from pyxel.calibration import Calibration
from pyxel.calibration.fitting import ModelFitting
from pyxel.calibration.util import CalibrationMode, ResultType
from pyxel.detectors import CCD
from pyxel.inputs_outputs import Configuration
from pyxel.pipelines import DetectionPipeline, Processor

try:
    import pygmo as pg

    WITH_PYGMO = True
except ImportError:
    WITH_PYGMO = False


def configure(mf: ModelFitting, sim: Configuration) -> None:
    """TBW."""
    assert sim.calibration is not None

    pg.set_global_rng_seed(sim.calibration.seed)
    np.random.seed(sim.calibration.seed)

    mf.configure(
        calibration_mode=sim.calibration.calibration_mode,
        generations=sim.calibration.algorithm.generations,
        population_size=sim.calibration.algorithm.population_size,
        simulation_output=sim.calibration.result_type,
        fitness_func=sim.calibration.fitness_function,
        target_output=sim.calibration.target_data_path,
        target_fit_range=sim.calibration.target_fit_range,
        out_fit_range=sim.calibration.result_fit_range,
        weighting=sim.calibration.weighting_path,
        file_path=None,
    )


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize("yaml_file", ["tests/data/calibrate.yaml"])
def test_configure_params(yaml_file):
    """Test """
    cfg = io.load(yaml_file)
    detector = cfg.ccd_detector
    pipeline = cfg.pipeline
    processor = Processor(detector, pipeline)
    calibration = cfg.calibration
    mf = ModelFitting(processor, calibration.parameters)

    if not isinstance(mf.processor, Processor):
        raise ValueError

    configure(mf, cfg)

    assert mf.calibration_mode == CalibrationMode.Pipeline
    assert mf.sim_fit_range == slice(2, 5, None)
    assert mf.targ_fit_range == slice(1, 4, None)
    assert mf.sim_output == ResultType.Image


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize("yaml", ["tests/data/calibrate_fits.yaml"])
def test_configure_fits_target(yaml):
    """Test """
    cfg = io.load(yaml)
    detector = cfg.ccd_detector
    pipeline = cfg.pipeline
    processor = Processor(detector, pipeline)
    calibration = cfg.calibration
    mf = ModelFitting(processor, calibration.parameters)
    configure(mf, cfg)
    assert mf.sim_fit_range == (slice(2, 5, None), slice(4, 7, None))
    assert mf.targ_fit_range == (slice(1, 4, None), slice(5, 8, None))
    assert mf.sim_output == ResultType.Image
    expected = np.array(
        [
            [3858.44799859, 3836.11204939, 3809.85008514],
            [4100.87410744, 4053.26348117, 4018.33656962],
            [4233.53215652, 4021.60164244, 3969.79740826],
        ]
    )
    np.testing.assert_array_equal(
        np.around(mf.all_target_data[0], decimals=4), np.around(expected, decimals=4)
    )


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "yaml", ["tests/data/calibrate.yaml", "tests/data/calibrate_fits.yaml"]
)
def test_boundaries(yaml):
    """Test """
    cfg = io.load(yaml)
    detector = cfg.ccd_detector
    pipeline = cfg.pipeline
    processor = Processor(detector, pipeline)
    calibration = cfg.calibration
    mf = ModelFitting(processor, calibration.parameters)

    configure(mf, cfg)

    lbd_expected = [1.0, -3.0, -3.0, -2.0, -2.0, 0.0, 10.0]
    ubd_expected = [10.0, 0.3010299956639812, 0.3010299956639812, 1.0, 1.0, 1.0, 200.0]
    assert mf.lbd == lbd_expected
    assert mf.ubd == ubd_expected
    ll, uu = mf.get_bounds()
    assert ll == lbd_expected
    assert uu == ubd_expected


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "simulated_data, target_data, expected_fitness",
    [
        (231, 231, 0.0),
        (231, 145, 86.0),
        (2.31, 1.45, 0.8600000000000001),
        (2.0, 1, 1.0),
        (np.array([1, 9, 45, 548, 2, 2]), np.array([1, 9, 45, 548, 2, 2]), 0.0),
        (np.array([1, 9, 45, 548, 2, 2]), np.array([1, 3, 56, 21, 235, 11]), 786.0),
        (
            np.array([[1362.0, 1378.0], [1308.0, 1309.0]]),
            np.array([[1362.0, 1378.0], [1308.0, 1309.0]]),
            0.0,
        ),
        (
            np.array([[1362.0, 1378.0], [1308.0, 1309.0]]),
            np.array([[1462.0, 1368.0], [1508.0, 1399.0]]),
            400.0,
        ),
    ],
)
def test_calculate_fitness(simulated_data, target_data, expected_fitness):
    """Test"""
    cfg = io.load("tests/data/calibrate.yaml")
    detector = cfg.ccd_detector
    pipeline = cfg.pipeline
    processor = Processor(detector, pipeline)
    calibration = cfg.calibration
    mf = ModelFitting(processor, calibration.parameters)
    configure(mf, cfg)
    fitness = mf.calculate_fitness(simulated_data, target_data)
    assert fitness == expected_fitness
    print("fitness: ", fitness)


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "yaml, factor, expected_fitness",
    [
        (Path("tests/data/calibrate_weighting.yaml"), 1, 0.0),
        (Path("tests/data/calibrate_weighting.yaml"), 2, 310815803081.51117),
        (Path("tests/data/calibrate_weighting.yaml"), 3, 621631606163.0223),
    ],
)
def test_weighting(yaml, factor, expected_fitness):
    """Test"""
    cfg = io.load(yaml)
    detector = cfg.ccd_detector
    pipeline = cfg.pipeline
    processor = Processor(detector, pipeline)
    calibration = cfg.calibration
    mf = ModelFitting(processor, calibration.parameters)
    configure(mf, cfg)
    fitness = mf.calculate_fitness(
        mf.all_target_data[0] * factor, mf.all_target_data[0]
    )
    assert fitness == expected_fitness
    print("fitness: ", fitness)


def custom_fitness_func(simulated, target, weighting=None):
    """Custom fitness func for testing"""
    return np.sum(target * 2 - simulated / 2 + 1.0)


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "yaml, simulated, target, weighting",
    [
        ("tests/data/calibrate_custom_fitness.yaml", 1.0, 2.0, 4.5),
        (
            "tests/data/calibrate_custom_fitness.yaml",
            np.array([1.0, 2.0, 3.0]),
            np.array([2.0, 5.0, 6.0]),
            26.0,
        ),
        ("tests/data/calibrate_least_squares.yaml", 2.0, 4.0, 4.0),
    ],
)
def test_custom_fitness(yaml, simulated, target, weighting):
    """Test"""
    cfg = io.load(yaml)
    assert isinstance(cfg, Configuration)

    detector = cfg.ccd_detector
    assert isinstance(detector, CCD)

    pipeline = cfg.pipeline
    assert isinstance(pipeline, DetectionPipeline)

    processor = Processor(detector, pipeline)
    assert isinstance(processor, Processor)

    calibration = cfg.calibration
    assert isinstance(calibration, Calibration)

    assert cfg.calibration is not None

    mf = ModelFitting(processor, variables=calibration.parameters)
    configure(mf=mf, sim=cfg)

    fitness = mf.calculate_fitness(simulated, target)
    assert fitness == weighting
    print("fitness: ", fitness)


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "yaml, parameter, expected_fitness",
    [
        (
            "tests/data/calibrate_models.yaml",
            np.array(
                [
                    1.0,
                    0.5,
                    1.5,
                    -2.0,
                    -3.0,
                    4.5,
                    -4.0,
                    1.0,
                    0.5,
                    -3.5,
                    2.0,
                    -3.0,
                    -4.0,
                    0.5,
                    1.0,
                    100.0,
                ]
            ),
            88663.74498686477,
        )
    ],
)
def test_fitness(yaml, parameter, expected_fitness):
    """Test"""
    cfg = io.load(yaml)

    detector = cfg.ccd_detector
    pipeline = cfg.pipeline

    processor = Processor(detector, pipeline)
    calibration = cfg.calibration

    mf = ModelFitting(processor, calibration.parameters)

    configure(mf, cfg)

    overall_fitness = mf.fitness(parameter)
    assert overall_fitness[0] == expected_fitness

    print("fitness: ", overall_fitness[0])


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "yaml, parameter, expected_array",
    [
        (
            "tests/data/calibrate_models.yaml",
            np.arange(16.0),
            np.array(
                [
                    0.0,
                    10.0,
                    100.0,
                    1000.0,
                    10000.0,
                    5.0,
                    6.0,
                    7.0,
                    8.0,
                    1.0e09,
                    1.0e10,
                    1.0e11,
                    1.0e12,
                    13.0,
                    1.0e14,
                    15.0,
                ]
            ),
        )
    ],
)
def test_split_and_update(yaml, parameter, expected_array):
    """Test"""
    cfg = io.load(yaml)
    detector = cfg.ccd_detector
    pipeline = cfg.pipeline
    processor = Processor(detector, pipeline)
    calibration = cfg.calibration
    mf = ModelFitting(processor, calibration.parameters)
    configure(mf, cfg)
    array = mf.convert_to_parameters(parameter)
    np.testing.assert_array_equal(array, expected_array)


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "yaml, param_array",
    [
        (
            "tests/data/calibrate_models.yaml",
            np.array(
                [
                    1.0,
                    10.0,
                    100.0,
                    1000.0,
                    10000.0,
                    5.0,
                    6.0,
                    7.0,
                    8.0,
                    1.0e09,
                    1.0e10,
                    1.0e11,
                    1.0e12,
                    13.0,
                    1.0e14,
                    150,
                ]
            ),
        )
    ],
)
def test_detector_and_model_update(yaml: str, param_array: np.ndarray):
    """Test"""
    cfg = io.load(yaml)

    detector = cfg.ccd_detector
    pipeline = cfg.pipeline

    processor = Processor(detector, pipeline)
    calibration = cfg.calibration
    mf = ModelFitting(processor=processor, variables=calibration.parameters)
    configure(mf=mf, sim=cfg)
    mf.processor = mf.update_processor(param_array, processor)

    assert mf.processor.pipeline.charge_transfer is not None
    assert mf.processor.pipeline.charge_measurement is not None

    attributes = [
        mf.processor.detector.characteristics.amp,
        mf.processor.pipeline.charge_transfer.models[0].arguments["tr_p"],
        mf.processor.pipeline.charge_transfer.models[0].arguments["nt_p"],
        mf.processor.pipeline.charge_transfer.models[0].arguments["sigma_p"],
        mf.processor.pipeline.charge_transfer.models[0].arguments["beta_p"],
        mf.processor.pipeline.charge_measurement.models[1].arguments["std_deviation"],
        mf.processor.detector.environment.temperature,
    ]
    a = 0
    for idx, attr in enumerate(attributes):
        if isinstance(attr, np.ndarray):
            b = len(attr)
            np.testing.assert_array_equal(attr, param_array[a : a + b])
        else:
            b = 1
            assert attr == param_array[a]
        a += b



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Simple model to load charge profiles."""
import logging
import typing as t
from pathlib import Path

import numpy as np

from pyxel.detectors import Detector


# TODO: Fix this
# @validators.validate
# @config.argument(name='txt_file', label='file path', units='', validate=checkers.check_path)
def charge_profile(
    detector: Detector,
    txt_file: str,
    fit_profile_to_det: bool = False,
    profile_position: t.Optional[list] = None,
) -> None:
    """Load charge profile from txt file for detector, mostly for but not limited to CCDs.

    :param detector: Pyxel Detector object
    :param txt_file: file path
    :param fit_profile_to_det: bool
    :param profile_position: list
    """
    logging.info("")
    geo = detector.geometry

    full_path = Path(txt_file).resolve()

    # All pixels has zero charge by default
    detector_charge = np.zeros((geo.row, geo.col))
    # Load 2d charge profile (which can be smaller or larger in dimensions than detector imaging area)
    charge_from_file = np.loadtxt(full_path, ndmin=2)
    if fit_profile_to_det:
        # Crop 2d charge profile, so it is not larger in dimensions than detector imaging area)
        charge_from_file = charge_from_file[slice(0, geo.row), slice(0, geo.col)]
    profile_rows, profile_cols = charge_from_file.shape
    if profile_position is None:
        profile_position = [0, 0]
    detector_charge[
        slice(profile_position[0], profile_position[0] + profile_rows),
        slice(profile_position[1], profile_position[1] + profile_cols),
    ] = charge_from_file
    charge_number = detector_charge.flatten()
    where_non_zero = np.where(charge_number > 0.0)
    charge_number = charge_number[where_non_zero]
    size = charge_number.size

    init_ver_pix_position = geo.vertical_pixel_center_pos_list()[where_non_zero]
    init_hor_pix_position = geo.horizontal_pixel_center_pos_list()[where_non_zero]

    detector.charge.add_charge(
        particle_type="e",
        particles_per_cluster=charge_number,
        init_energy=[0.0] * size,
        init_ver_position=init_ver_pix_position,
        init_hor_position=init_hor_pix_position,
        init_z_position=[0.0] * size,
        init_ver_velocity=[0.0] * size,
        init_hor_velocity=[0.0] * size,
        init_z_velocity=[0.0] * size,
    )



#   --------------------------------------------------------------------------
#   Copyright 2019 SCI-FIV, ESA (European Space Agency)
#   Written by B. Serra
#   --------------------------------------------------------------------------
"""Pyxel persistence model."""
import logging

import numpy as np

from pyxel.detectors import CMOS


# @validators.validate
# @config.argument(name='', label='', units='', validate=)
def simple_persistence(
    detector: CMOS, trap_timeconstants: list, trap_densities: list
) -> None:
    """Trapping/detrapping charges."""
    logging.info("Persistence")
    if "persistence" not in detector._memory.keys():
        detector._memory["persistence"] = dict()
        for trap_density, trap_timeconstant in zip(trap_densities, trap_timeconstants):
            entry = "".join(
                ["trappedCharges_", str(trap_density), "-", str(trap_timeconstant)]
            )
            detector._memory["persistence"].update(
                {entry: np.zeros((detector.geometry.row, detector.geometry.col))}
            )
    else:
        for trap_density, trap_timeconstant in zip(trap_densities, trap_timeconstants):
            entry = "".join(
                ["trappedCharges_", str(trap_density), "-", str(trap_timeconstant)]
            )
            trapped_charges = detector._memory["persistence"][entry]
            # Trap density is a scalar for now, in the future we could feed maps?
            trapped_charges = trapped_charges + (0.5 / trap_timeconstant) * (
                detector.pixel.array * trap_density - trapped_charges
            )
            # Remove the trapped charges from the pixel
            detector.pixel.array -= trapped_charges.astype(np.int32)
            # Replace old trapped charges map in the detector's memory
            detector._memory["persistence"][entry] = trapped_charges



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Classes for creating outputs."""
import typing as t
from enum import Enum
from glob import glob
from pathlib import Path

import attr
from matplotlib import pyplot as plt
from matplotlib.ticker import ScalarFormatter


@attr.s(auto_attribs=True, slots=True, frozen=True)
class PlotArguments:
    """TBW."""

    title: t.Optional[str] = None
    xscale: str = "linear"
    yscale: str = "linear"
    xlabel: t.Optional[str] = None
    ylabel: t.Optional[str] = None
    xlim: t.Tuple[t.Optional[float], t.Optional[float]] = (None, None)
    ylim: t.Tuple[t.Optional[float], t.Optional[float]] = (None, None)
    xticks: t.Any = None
    yticks: t.Any = None
    sci_x: bool = False
    sci_y: bool = False
    grid: bool = False
    axis: t.Any = None
    bins: t.Optional[int] = None  # TODO: This should not be here !

    @classmethod
    def from_dict(cls, dct: t.Mapping[str, t.Any]) -> "PlotArguments":
        """TBW."""
        return cls(**dct)

    def to_dict(self) -> t.Dict[str, t.Any]:
        """TBW."""
        return attr.asdict(self)


class PlotType(Enum):
    """TBW."""

    Histogram = "histogram"
    Graph = "graph"
    Scatter = "scatter"


def show_plots() -> None:
    """Close last empty canvas and show all the previously created figures."""
    plt.close()
    plt.show()


def update_plot(ax_args: dict, ax: plt.Axes) -> None:
    """TBW.

    Parameters
    ----------
    ax_args
    ax
    """
    ax.set_xlabel(ax_args["xlabel"])
    ax.set_ylabel(ax_args["ylabel"])

    ax.set_xscale(ax_args["xscale"])
    ax.set_yscale(ax_args["yscale"])

    ax.set_xlim(ax_args["xlim"][0], ax_args["xlim"][1])
    ax.set_ylim(ax_args["ylim"][0], ax_args["ylim"][1])

    ax.set_title(ax_args["title"])

    if ax_args["axis"]:
        # TODO: Fix this
        raise NotImplementedError
        # plt.axis(ax_args["axis"])

    if ax_args["xticks"]:
        ax.set_xticks(ax_args["xticks"])
    if ax_args["yticks"]:
        ax.set_yticks(ax_args["yticks"])

    # TODO: Enable this
    ax.grid(ax_args["grid"])

    if ax_args["sci_x"]:
        ax.xaxis.set_major_formatter(ScalarFormatter())
        ax.ticklabel_format(style="scientific", axis="x", scilimits=(0, 0))
    if ax_args["sci_y"]:
        ax.yaxis.set_major_formatter(ScalarFormatter())
        ax.ticklabel_format(style="scientific", axis="y", scilimits=(0, 0))


# TODO: Refactor this function
def update_fits_header(
    header: dict, key: t.Union[str, list, tuple], value: t.Any
) -> None:
    """TBW.

    Parameters
    ----------
    header
    key
    value
    """
    if isinstance(value, (str, int, float)):
        result = value  # type: t.Union[str, int, float]
    else:
        result = repr(value)

    if isinstance(result, str):
        result = result[0:24]

    if isinstance(key, (list, tuple)):
        key = "/".join(key)

    key = key.replace(".", "/")[0:36]
    header[key] = value


# TODO: Create unit tests
# TODO: Refactor this in 'def apply_run_number(folder, template_filename) -> Path
def apply_run_number(template_filename: Path) -> Path:
    """Convert the file name numeric placeholder to a unique number.

    :param template_filename:
    :return:
    """
    path_str = str(template_filename)
    if "?" in path_str:
        # TODO: Use method 'Path.glob'
        dir_list = sorted(glob(path_str))

        p_0 = path_str.find("?")
        p_1 = path_str.rfind("?")
        template = path_str[slice(p_0, p_1 + 1)]
        path_str = path_str.replace(template, "{:0%dd}" % len(template))

        last_num = 0
        if len(dir_list):
            path_last = dir_list[-1]
            last_num = int(path_last[slice(p_0, p_1 + 1)])
        last_num += 1
        path_str = path_str.format(last_num)

    return Path(path_str)


# TODO: the log file should directly write in 'output_dir'
def save_log_file(output_dir: Path) -> None:
    """Move log file to the outputs directory of the simulation."""
    log_file = Path("pyxel.log").resolve(strict=True)  # type: Path

    new_log_filename = output_dir.joinpath(log_file.name)
    log_file.rename(new_log_filename)



from collections import abc
from pathlib import Path

import pytest

from pyxel import inputs_outputs as io
from pyxel.detectors import CCD
from pyxel.inputs_outputs import Configuration
from pyxel.parametric.parametric import Parametric, ParametricMode
from pyxel.pipelines import DetectionPipeline, Processor

try:
    import pygmo as pg

    WITH_PYGMO = True
except ImportError:
    WITH_PYGMO = False


expected_sequential = [
    (0, [("level", 10), ("initial_energy", 100)]),
    (1, [("level", 20), ("initial_energy", 100)]),
    (2, [("level", 30), ("initial_energy", 100)]),
    (3, [("level", 100), ("initial_energy", 100)]),
    (4, [("level", 100), ("initial_energy", 200)]),
    (5, [("level", 100), ("initial_energy", 300)]),
]

expected_embedded = [
    (0, [("level", 10), ("initial_energy", 100)]),
    (1, [("level", 10), ("initial_energy", 200)]),
    (2, [("level", 10), ("initial_energy", 300)]),
    (3, [("level", 20), ("initial_energy", 100)]),
    (4, [("level", 20), ("initial_energy", 200)]),
    (5, [("level", 20), ("initial_energy", 300)]),
    (6, [("level", 30), ("initial_energy", 100)]),
    (7, [("level", 30), ("initial_energy", 200)]),
    (8, [("level", 30), ("initial_energy", 300)]),
]


@pytest.mark.skipif(not WITH_PYGMO, reason="Package 'pygmo' is not installed.")
@pytest.mark.parametrize(
    "mode, expected",
    [
        # ('single', expected_single),
        (ParametricMode.Sequential, expected_sequential),
        (ParametricMode.Embedded, expected_embedded),
    ],
)
def test_pipeline_parametric_without_init_photon(mode: ParametricMode, expected):
    input_filename = "tests/data/parametric.yaml"
    cfg = io.load(Path(input_filename))

    assert isinstance(cfg, Configuration)
    assert hasattr(cfg, "parametric")
    assert hasattr(cfg, "ccd_detector")
    assert hasattr(cfg, "pipeline")

    parametric = cfg.parametric
    assert isinstance(parametric, Parametric)

    parametric.parametric_mode = mode

    detector = cfg.ccd_detector
    assert isinstance(detector, CCD)

    assert detector.has_photon is False
    with pytest.raises(
        RuntimeError,
        match=(
            r"Photon array is not initialized ! "
            r"Please use a 'Photon Generation' model"
        ),
    ):
        _ = detector.photon

    pipeline = cfg.pipeline
    assert isinstance(pipeline, DetectionPipeline)

    processor = Processor(
        detector=detector, pipeline=pipeline
    )  # type: pyxel.pipelines.processor.Processor
    result = parametric.debug(processor)
    assert result == expected

    configs = parametric.collect(processor)
    assert isinstance(configs, abc.Iterator)

    for config in configs:
        assert isinstance(config, Processor)

        config.run_pipeline()



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Readout electronics model."""
import logging
from pydoc import locate

import numpy as np

from pyxel.detectors import Detector

# from astropy import units as u


# TODO: Fix this
# @validators.validate
# @config.argument(name='data_type', label='type of output data array', units='ADU',
#                  validate=checkers.check_choices(['numpy.uint16', 'numpy.uint32', 'numpy.uint64',
#                                                   'numpy.int32', 'numpy.int64']))
def simple_digitization(detector: Detector, data_type: str = "numpy.uint16") -> None:
    """Digitize signal array mimicking readout electronics.

    :param detector: Pyxel Detector object
    :param data_type: numpy integer type: ``numpy.uint16``, ``numpy.uint32``, ``numpy.uint64``,
                                          ``numpy.int32``, ``numpy.int64``
    """
    logging.info("")

    d_type = locate(data_type)
    if d_type is None:
        raise TypeError(
            "Can not locate the type defined as `data_type` argument in yaml file."
        )
    # Gain of the Analog-Digital Converter
    detector.signal.array *= detector.characteristics.a2
    # floor of signal values element-wise (quantization)
    detector.signal.array = np.floor(detector.signal.array)
    # convert floats to other datatype (e.g. 16-bit unsigned integers)
    detector.signal.array = np.clip(
        detector.signal.array, a_min=np.iinfo(d_type).min, a_max=np.iinfo(d_type).max
    )
    detector.image.array = detector.signal.array.astype(d_type)


def simple_processing(detector: Detector) -> None:
    """Create an image array from signal array.

    :param detector: Pyxel Detector object
    """
    logging.info("")
    detector.signal.array *= detector.characteristics.a2
    detector.image.array = detector.signal.array


def sar_adc(detector: Detector, adc_bits: int = 16, range_volt: tuple = (0, 5)) -> None:
    """Digitize signal array using SAR ADC logic.

    :param detector: Pyxel Detector object
    :param adc_bits: integer: Number of bits for the ADC
    :param range_volt: tuple with min anx max volt value
    """
    # import numpy as np
    logging.info("")
    # Extract the data to digitize
    data = detector.signal.array
    data_digitized = np.zeros((detector.geometry.row, detector.geometry.col))
    # First normalize the data to voltage since there is no model for
    # the conversion photogenerated carrier > volts in the model/charge_measure
    data = data * range_volt[1] / np.max(data)
    # steps = list()
    # Set the reference voltage of the ADC to half the max
    ref = range_volt[1] / 2.0
    # For each bits, compare the value of the ref to the capacitance value
    for i in np.arange(0, adc_bits):
        # digital value associated with this step
        digital_value = 2 ** (adc_bits - (i + 1))
        # All data that is higher than the ref is equal to the dig. value
        data_digitized[data >= ref] += digital_value
        # Subtract ref value from the data
        data[data >= ref] -= ref
        # steps.append(ref)
        # Divide reference voltage by 2 for next step
        ref /= 2.0

    detector.image.array = data_digitized
    # return data_digitized, steps



"""Functional tests for the Pyxel pipelines."""
#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

# import os
# import shutil
# import pytest
# import numpy as np
# from astropy.io import fits
# from pyxel.run import run


# @pytest.mark.parametrize('yaml, expected_image, seed, outfile',
#                          [
#                              ('tests/data/photon_transfer_function.yaml', 'tests/data/uniform_1000.fits', 1111,
#                               'run_01/image_01.fits'),
#                              ('tests/data/pipeline_01.yaml', 'tests/data/expected_pipeline_01.fits', 1111,
#                               'run_02/image_01.fits'),
#                              ('tests/data/pipeline_02.yaml', 'tests/data/expected_pipeline_01.fits', 1111,
#                               'run_03/image_01.fits'),
#                          ])
# def test_pyxel_pipeline(yaml, expected_image, seed, outfile):
#     """Test """
#     outdir = 'tests/data'
#     if os.path.exists(outdir+'/'+outfile[:6]):
#         shutil.rmtree(outdir+'/'+outfile[:6])      # TODO: THIS DOES NOT WORK FOR 2ND TIME!!!!
#
#     run(input_filename=yaml, output_directory=outdir, random_seed=seed)
#     image = fits.getdata(outdir+'/'+outfile)
#     expected_image = fits.getdata(expected_image)
#     np.testing.assert_array_equal(image, expected_image)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW."""
import functools
import inspect
import typing as t

if t.TYPE_CHECKING:
    from pyxel.detectors import Detector


# Declare type variable
T = t.TypeVar("T")


# TODO: Improve this class. See issue #133.
class Arguments(dict):
    """TBW."""

    def __getattr__(self, name: str) -> t.Union[int, float]:
        if name not in self:
            raise AttributeError(f"Argument {name!r} does not exist.")

        result = self[name]
        assert isinstance(result, int) or isinstance(result, float)

        return result

    def __dir__(self):
        return dir(type(self)) + list(self)

    # def __deepcopy__(self, memo) -> "Arguments":
    #     """TBW."""
    #     return Arguments(deepcopy(self._arguments))


# TODO: Improve this class. See issue #132.
class ModelFunction:
    """Create a wrapper function around a Model function.

    Examples
    --------
    >>> from pyxel.models.photon_generation.illumination import illumination
    >>> model_func = ModelFunction(
    ...     func=illumination,
    ...     name="illumination",
    ...     arguments={"level": 1, "option": "foo"},
    ... )

    Access basic parameters
    >>> model_func.name
    'illumination'
    >>> model_func.enabled
    True
    >>> model_func.arguments
    Arguments({'level': 1, 'option': 'foo'})

    Access the arguments with a ``dict`` interface
    >>> list(model_func.arguments)
    ['level', 'option']
    >>> model_func.arguments["level"]
    1
    >>> model_func.arguments["level"] = 2
    TypeError: 'Arguments' object does not support item assignment

    Access the arguments with an attribute interface
    >>> model_func.arguments.level
    1
    """

    def __init__(
        self,
        func: t.Callable,
        name: str,
        arguments: t.Optional[dict] = None,
        enabled: bool = True,
    ):
        assert not inspect.isclass(func)

        self._func = func  # type: t.Callable
        self._name = name
        self.enabled = enabled  # type: bool

        if arguments is None:
            arguments = {}

        self._arguments = Arguments(arguments)
        # self.group = None               # TODO

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__  # type: str
        func_name = self._func.__module__ + "." + self._func.__name__

        return (
            f"{cls_name}(name={self.name!r}, func={func_name}, "
            f"arguments={self.arguments!r}, enabled={self.enabled!r})"
        )

    @property
    def name(self) -> str:
        """TBW."""
        return self._name

    @property
    def arguments(self) -> dict:
        """TBW."""
        return self._arguments

    def change_argument(self, argument: str, value: t.Any) -> None:
        """Change a model argument.

        Parameters
        ----------
        argument: str
            Name of the argument to be changed.
        value

        Returns
        -------
        None

        Raises
        ------
        TypeError
            If types of the changed value and default value do not match.
        KeyError
            If argument does not exist.
        """
        try:
            if type(self._arguments[argument]) == type(value):
                self._arguments[argument] = value
            else:
                raise TypeError(
                    f"Type of the changed value should be {type(self._arguments[argument])}, not {type(value)}"
                )
        except KeyError:
            raise

    # # TODO: Replace this by __call__ ?
    # @property
    # def function(self) -> t.Callable:
    #     """TBW."""
    #     func_ref = evaluate_reference(self.func)  # type: t.Callable
    #     if isinstance(func_ref, type):
    #         # this is a class type, instantiate it using default arguments.
    #         func_ref = func_ref()
    #         # TODO: should check whether or not it's callable.
    #     func = functools.partial(func_ref, **self.arguments)
    #     return func

    def __call__(self, detector: "Detector") -> T:
        """TBW."""
        # func_ref = evaluate_reference(self.func)  # type: t.Callable

        # if inspect.isclass(self._func):
        #     # this is a class type, instantiate it using default arguments.
        #     func_ref = self._func()
        #     # TODO: should check whether or not it's callable.
        #     raise NotImplementedError
        #
        # else:
        #     func_ref = self._func

        func = functools.partial(self._func, **self.arguments)

        result = func(detector)  # type: T

        return result



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Detector class."""
import collections
import typing as t
from math import sqrt
from pathlib import Path

import numpy as np

from pyxel.data_structure import Charge, Image, Photon, Pixel, Signal
from pyxel.detectors import Environment, Material
from pyxel.util.memory import get_size, memory_usage_details

__all__ = ["Detector"]


# TODO: Add methods to save/load a `Detector` instance to the filesystem
#       Example of methods:
#           def to_fits(self, filename: Path):      # Save into one FITS file that contains multiple HDUs
#               ...
#           @classmethod
#           def from_fits(self, filename: Path) -> Detector     # Store into one FITS file
#               ...
#           def to_hdf5(...) / def from_hdf5(...)
#           def to_folder(...)  / def from_folder(...)
#           def to_yaml(...) ? / def from_yaml(...)
#           def to_asdf(...) ? / def from_asdf(...)
class Detector:
    """The detector class."""

    def __init__(self, material: Material, environment: Environment):
        self.material = material  # type: Material
        self.environment = environment  # type: Environment

        self.header = collections.OrderedDict()  # type: t.Dict[str, object]

        self._photon = None  # type: t.Optional[Photon]
        self._charge = None  # type: t.Optional[Charge]
        self._pixel = None  # type: t.Optional[Pixel]
        self._signal = None  # type: t.Optional[Signal]
        self._image = None  # type: t.Optional[Image]

        # This will be the memory of the detector where trapped charges will be saved
        self._memory = dict()  # type: t.Dict

        self.input_image = None  # type: t.Optional[np.ndarray]
        self._output_dir = None  # type: t.Optional[Path]  # TODO: Is it really needed ?

        self.start_time = 0.0  # type: float
        self.end_time = 0.0  # type: float
        self.steps = 0  # type: int
        self.time_step = 0.0  # type: float
        self._time = 0.0  # type: float
        self._dynamic = False  # type: bool
        self._non_destructive = False  # type: bool
        self.read_out = True  # type: bool
        self._all_time_steps_it = iter([])  # type: t.Iterator[float]

        self._numbytes = 0

    @property
    def geometry(self):
        """TBW."""
        raise NotImplementedError

    @property
    def characteristics(self):
        """TBW."""
        raise NotImplementedError

    @property
    def has_photon(self) -> bool:
        """TBW."""
        return self._photon is not None

    @property
    def photon(self) -> Photon:
        """TBW."""
        if not self._photon:
            raise RuntimeError(
                "Photon array is not initialized ! "
                "Please use a 'Photon Generation' model"
            )
        return self._photon

    @photon.setter
    def photon(self, obj: Photon) -> None:
        self._photon = obj

    @property
    def charge(self) -> Charge:
        """TBW."""
        if not self._charge:
            raise RuntimeError("'charge' not initialized.")

        return self._charge

    @property
    def pixel(self) -> Pixel:
        """TBW."""
        if not self._pixel:
            raise RuntimeError("'pixel' not initialized.")

        return self._pixel

    @property
    def signal(self) -> Signal:
        """TBW."""
        if not self._signal:
            raise RuntimeError("'signal' not initialized.")

        return self._signal

    @property
    def image(self) -> Image:
        """TBW."""
        if not self._image:
            raise RuntimeError("'image' not initialized.")

        return self._image

    # TODO: Rename to 'reset' ?
    def initialize(self, reset_all: bool = True) -> None:
        """TBW."""
        self._photon = None
        if reset_all:
            self._charge = Charge()
            self._pixel = Pixel(self.geometry)
            self._signal = Signal(self.geometry)
            self._image = Image(self.geometry)

    # TODO: Set an `Output` object ?
    # TODO: Is it really needed ?
    def set_output_dir(self, path: t.Union[str, Path]) -> None:
        """Set output directory path."""
        self._output_dir = Path(path)

    # TODO: Get an `Output object ?
    # TODO: Is it really needed ?
    @property
    def output_dir(self) -> Path:
        """Output directory path."""
        if self._output_dir is None:
            raise RuntimeError("'output_dir' is not initialized.")

        return self._output_dir

    def set_dynamic(
        self, time_step: float, steps: int, ndreadout: bool = False
    ) -> None:
        """Switch on dynamic (time dependent) mode."""
        self._dynamic = True
        self.time_step = time_step
        self.steps = steps
        self._non_destructive = ndreadout
        self.end_time = self.time_step * self.steps

        all_time_steps = np.round(
            np.linspace(self.time_step, self.end_time, self.steps, endpoint=True),
            decimals=10,
        )
        self._all_time_steps_it = map(float, all_time_steps)

    @property
    def is_dynamic(self) -> bool:
        """Return if detector is dynamic (time dependent) or not.

        By default it is not dynamic.
        """
        return self._dynamic

    @property
    def is_non_destructive_readout(self) -> bool:
        """Return if detector readout mode is destructive or integrating.

        By default it is destructive (non-integrating).
        """
        return self._non_destructive

    @property
    def e_thermal_velocity(self) -> float:
        """TBW.

        :return:
        """
        k_boltzmann = 1.38064852e-23  # J/K
        return sqrt(
            3
            * k_boltzmann
            * self.environment.temperature
            / self.material.e_effective_mass
        )

    @property
    def time(self) -> float:  # TODO
        """TBW."""
        return self._time

    # TODO: This method is used in 'run.py'. We could implement this as an iterator.
    def elapse_time(self) -> float:
        """TBW."""
        try:
            self._time = float(next(self._all_time_steps_it))
        except StopIteration:
            self._time = 0.0
        return self._time

    @property
    def numbytes(self) -> int:
        """Recursively calculates object size in bytes using Pympler library.

        Returns
        -------
        int
            Size of the object in bytes.
        """
        self._numbytes = get_size(self)
        return self._numbytes

    def memory_usage(
        self, print_result: bool = True, human_readable: bool = True
    ) -> dict:
        """TBW.

        Returns
        -------
        dict
            Dictionary of attribute memory usage
        """
        attributes = [
            "_photon",
            "_charge",
            "_pixel",
            "_signal",
            "_image",
            "material",
            "environment",
            "_geometry",
            "_characteristics",
        ]

        return memory_usage_details(
            self, attributes, print_result=print_result, human_readable=human_readable
        )
